# from libc.stdlib cimport malloc, free
from libc.stdint cimport int64_t, uintptr_t, uint32_t, int32_t, uint16_t, uint8_t, int8_t
from cpython.mem cimport PyMem_Malloc, PyMem_Free
# from cpython.view cimport array as cvarray
import numpy as np
cimport numpy as np
import sys

cdef extern from "pgenlib_python_support.h":
    # macros aren't namespaced
    uintptr_t DIV_UP(uintptr_t val, uintptr_t divisor)

cdef extern from "pgenlib_python_support.h" namespace "plink2":
    ctypedef uint32_t boolerr_t
    ctypedef enum pglerr_t:
        kPglRetSuccess
        kPglRetSkipped
        kPglRetNomem
        kPglRetOpenFail
        kPglRetReadFail
        kPglRetWriteFail
        kPglRetMalformedInput
        kPglRetInconsistentInput
        kPglRetInvalidCmdline
        kPglRetHelp
        kPglRetThreadCreateFail
        kPglRetNetworkFail
        kPglRetSampleMajorBed
        kPglRetImproperFunctionCall
        kPglRetNotYetSupported
        kPglRetLongLine
        kPglRetEmptyFile

    boolerr_t cachealigned_malloc(uintptr_t size, void* aligned_pp)
    void aligned_free(void* aligned_ptr)

    void fill_ulong_zero(uintptr_t entry_ct, uintptr_t* ularr)
    # void bitvec_and(const uintptr_t* arg_bitvec, uintptr_t word_ct, uintptr_t* main_bitvec)
    void fill_interleaved_mask_vec(const uintptr_t* subset_mask, uint32_t base_vec_ct, uintptr_t* interleaved_mask_vec)
    void fill_cumulative_popcounts(const uintptr_t* subset_mask, uint32_t word_ct, uint32_t* cumulative_popcounts)

    ctypedef uintptr_t vul_t
    void transpose_quaterblock(const uintptr_t* read_iter, uint32_t read_ul_stride, uint32_t write_ul_stride, uint32_t read_batch_size, uint32_t write_batch_size, uintptr_t* write_iter, vul_t* vecaligned_buf)
    void transpose_bitblock(const uintptr_t* read_iter, uint32_t read_ul_stride, uint32_t write_ul_stride, uint32_t read_batch_size, uint32_t write_batch_size, uintptr_t* write_iter, vul_t* vecaligned_buf)

    void genovec_invert_unsafe(uint32_t sample_ct, uintptr_t* genovec)
    void biallelic_dosage16_invert(uint32_t dosage_ct, uint16_t* dosage_vals)

    void genoarr_to_bytes_minus9(const uintptr_t* genoarr, uint32_t sample_ct, int8_t* genobytes)
    void genoarr_to_int32s_minus9(const uintptr_t* genoarr, uint32_t sample_ct, int32_t* geno_int32)
    void genoarr_to_int64s_minus9(const uintptr_t* genoarr, uint32_t sample_ct, int64_t* geno_int64)
    void genoarr_to_allele_codes(const uintptr_t* genoarr, uint32_t sample_ct, int32_t* allele_codes)
    void genoarr_phased_to_allele_codes(const uintptr_t* genoarr, const uintptr_t* phasepresent, const uintptr_t* phaseinfo, uint32_t sample_ct, uint32_t phasepresent_ct, unsigned char* phasebytes, int32_t* allele_codes)
    void genoarr_phased_to_hap_codes(const uintptr_t* genoarr, const uintptr_t* phaseinfo, uint32_t variant_batch_size, int32_t* hap0_codes_iter, int32_t* hap1_codes_iter)
    void dosage16_to_floats_minus9(const uintptr_t* genoarr, const uintptr_t* dosage_present, const uint16_t* dosage_vals, uint32_t sample_ct, uint32_t dosage_ct, float* geno_float)
    void dosage16_to_doubles_minus9(const uintptr_t* genoarr, const uintptr_t* dosage_present, const uint16_t* dosage_vals, uint32_t sample_ct, uint32_t dosage_ct, double* geno_double)
    void bytes_to_bits_unsafe(const uint8_t* boolbytes, uint32_t sample_ct, uintptr_t* bitarr)
    void bytes_to_genoarr_unsafe(const int8_t* genobytes, uint32_t sample_ct, uintptr_t* genoarr)
    void allele_codes_to_genoarr_unsafe(const int32_t* allele_codes, const unsigned char* phasepresent_bytes, uint32_t sample_ct, uintptr_t* genoarr, uintptr_t* phasepresent, uintptr_t* phaseinfo)
    void floats_to_dosage16(const float* floatarr, uint32_t sample_ct, uint32_t hard_call_halfdist, uintptr_t* genoarr, uintptr_t* dosage_present, uint16_t* dosage_vals, uint32_t* dosage_ct_ptr)
    void doubles_to_dosage16(const double* doublearr, uint32_t sample_ct, uint32_t hard_call_halfdist, uintptr_t* genoarr, uintptr_t* dosage_present, uint16_t* dosage_vals, uint32_t* dosage_ct_ptr)

    cdef enum:
        k1LU
    cdef enum:
        kCacheline
    cdef enum:
        kBitsPerWord
    cdef enum:
        kBitsPerWordD2
    cdef enum:
        kBitsPerVec
    cdef enum:
        kBitsPerCacheline
    cdef enum:
        kQuatersPerVec
    cdef enum:
        kQuatersPerCacheline
    cdef enum:
        kBytesPerVec
    cdef enum:
        kInt32PerVec
    cdef enum:
        kInt32PerCacheline
    cdef enum:
        kWordsPerVec
    cdef enum:
        kPglErrstrBufBlen

    cdef enum:
        kPglQuaterTransposeBatch
    cdef enum:
        kPglQuaterTransposeWords
    cdef enum:
        kPglQuaterTransposeBufbytes
    cdef enum:
        kPglBitTransposeBufbytes

    ctypedef uint32_t pgen_global_flags_t
    cdef enum:
        kfPgenGlobal0
    cdef enum:
        kfPgenGlobalHardcallPhasePresent
    cdef enum:
        kfPgenGlobalDosagePresent

    cdef cppclass pgen_file_info_t:
        uint32_t raw_variant_ct
        uint32_t raw_sample_ct        
        unsigned char* vrtypes
        uint32_t gflags

    ctypedef uint32_t pgen_header_ctrl_t

    void pgfi_preinit(pgen_file_info_t* pgfip)
    
    pglerr_t pgfi_init_phase1(const char* fname, uint32_t raw_variant_ct, uint32_t raw_sample_ct, uint32_t use_mmap, pgen_header_ctrl_t* header_ctrl_ptr, pgen_file_info_t* pgfip, uintptr_t* pgfi_alloc_cacheline_ct_ptr, char* errstr_buf)

    pglerr_t pgfi_init_phase2(pgen_header_ctrl_t header_ctrl, uint32_t allele_cts_already_loaded, uint32_t nonref_flags_already_loaded, uint32_t use_blockload, uint32_t vblock_idx_start, uint32_t vidx_end, uint32_t* max_vrec_width_ptr, pgen_file_info_t* pgfip, unsigned char* pgfi_alloc, uintptr_t* pgr_alloc_cacheline_ct_ptr, char* errstr_buf)
    
    cdef cppclass pgen_reader_t:
        pgen_file_info_t fi
        unsigned char* fread_buf

    void pgr_preinit(pgen_reader_t* pgrp)

    pglerr_t pgr_init(const char* fname, uint32_t max_vrec_width, pgen_file_info_t* pgfip, pgen_reader_t* pgrp, unsigned char* pgr_alloc)

    pglerr_t pgr_read_allele_countvec_subset_unsafe(const uintptr_t* sample_include, const uint32_t* sample_include_cumulative_popcounts, uint32_t sample_ct, uint32_t vidx, uint32_t allele_idx, pgen_reader_t* pgrp, uintptr_t* allele_countvec)

    pglerr_t pgr_read_refalt1_genovec_hphase_subset_unsafe(const uintptr_t* sample_include, const uint32_t* sample_include_cumulative_popcounts, uint32_t sample_ct, uint32_t vidx, pgen_reader_t* pgrp, uintptr_t* genovec, uintptr_t* phasepresent, uintptr_t* phaseinfo, uint32_t* phasepresent_ct_ptr)

    pglerr_t pgr_read_refalt1_genovec_dosage16_subset_unsafe(const uintptr_t* sample_include, const uint32_t* sample_include_cumulative_popcounts, uint32_t sample_ct, uint32_t vidx, pgen_reader_t* pgrp, uintptr_t* genovec, uintptr_t* dosage_present, uint16_t* dosage_vals, uint32_t* dosage_ct_ptr, uint32_t* is_explicit_alt1_ptr)
    
    pglerr_t pgr_get_refalt1_genotype_counts(const uintptr_t* sample_include, const uintptr_t* sample_include_interleaved_vec, const uint32_t* sample_include_cumulative_popcounts, uint32_t sample_ct, uint32_t vidx, pgen_reader_t* pgrp, uint32_t* genocounts)
    
    boolerr_t pgfi_cleanup(pgen_file_info_t* pgfip)
    boolerr_t pgr_cleanup(pgen_reader_t* pgrp)

    cdef cppclass pgen_writer_common_t:
        uint32_t variant_ct
        uint32_t sample_ct
        uintptr_t* allele_idx_offsets
        uint32_t vidx

    cdef cppclass st_pgen_writer_t:
        pgen_writer_common_t pwc

    pglerr_t spgw_init_phase1(const char* fname, uintptr_t* allele_idx_offsets, uintptr_t* explicit_nonref_flags, uint32_t variant_ct, uint32_t sample_ct, pgen_global_flags_t phase_dosage_gflags, uint32_t nonref_flags_storage, st_pgen_writer_t* spgwp, uintptr_t* alloc_cacheline_ct_ptr, uint32_t* max_vrec_len_ptr)

    void spgw_init_phase2(uint32_t max_vrec_len, st_pgen_writer_t* spgwp, unsigned char* spgw_alloc)

    pglerr_t spgw_append_biallelic_genovec(const uintptr_t* genovec, st_pgen_writer_t* spgwp)
    
    pglerr_t spgw_append_biallelic_genovec_hphase(const uintptr_t* genovec, const uintptr_t* phasepresent, const uintptr_t* phaseinfo, st_pgen_writer_t* spgwp)

    pglerr_t spgw_append_biallelic_genovec_dosage16(const uintptr_t* genovec, const uintptr_t* dosage_present, const uint16_t* dosage_vals, uint32_t dosage_ct, st_pgen_writer_t* spgwp)
    
    pglerr_t spgw_finish(st_pgen_writer_t* spgwp)
    
    boolerr_t spgw_cleanup(st_pgen_writer_t* spgwp)

    
cdef class PgenReader:
    # todo: nonref_flags, multiallelic variant support
    cdef pgen_file_info_t* _info_ptr
    cdef pgen_reader_t* _state_ptr
    cdef uintptr_t* _subset_include_vec
    cdef uintptr_t* _subset_include_interleaved_vec
    cdef uint32_t* _subset_cumulative_popcounts
    cdef uint32_t _subset_size
    # preallocate buffers we'll use repeatedly
    cdef uintptr_t* _genovec
    cdef uintptr_t* _phasepresent
    cdef uintptr_t* _phaseinfo
    cdef uintptr_t* _dosage_present
    cdef uint16_t* _dosage_vals
    cdef vul_t* _transpose_batch_buf
    # for multi-variant load-and-transpose, we load up to
    # kPglQuaterTransposeBatch (= 256) variants at a time, and then transpose
    cdef uintptr_t* _multivar_vmaj_geno_buf
    cdef uintptr_t* _multivar_vmaj_phasepresent_buf
    cdef uintptr_t* _multivar_vmaj_phaseinfo_buf
    cdef uintptr_t* _multivar_smaj_geno_batch_buf
    cdef uintptr_t* _multivar_smaj_phaseinfo_batch_buf
    cdef uintptr_t* _multivar_smaj_phasepresent_batch_buf

    cdef set_sample_subset_internal(self, np.ndarray[np.uint32_t,mode="c",ndim=1] sample_subset):
        cdef uint32_t raw_sample_ct = self._info_ptr[0].raw_sample_ct
        cdef uint32_t raw_sample_ctv = DIV_UP(raw_sample_ct, kBitsPerVec)
        cdef uint32_t raw_sample_ctaw = raw_sample_ctv * kWordsPerVec
        cdef uintptr_t* sample_include = self._subset_include_vec
        fill_ulong_zero(raw_sample_ctaw, sample_include)
        cdef uint32_t subset_size = sample_subset.size
        if subset_size == 0:
            raise RuntimeError("Empty sample_subset is not currently permitted.")
        cdef sample_uidx = sample_subset[0]
        cdef uint32_t idx = 0
        cdef next_uidx
        while True:
            if sample_uidx >= raw_sample_ct:
                raise RuntimeError("0-based sample idx too large (" + str(sample_uidx) + "; only " + str(raw_sample_ct) + " in file).")
            sample_include[sample_uidx / kBitsPerWord] |= k1LU << (sample_uidx % kBitsPerWord)
            idx += 1
            if idx == subset_size:
                break
            next_uidx = sample_subset[idx]

            # prohibit this since it implies that the caller expects genotypes
            # to be returned in a different order
            if next_uidx <= sample_uidx:
                raise RuntimeError("sample_subset is not in strictly increasing order.")
            
            sample_uidx = next_uidx

        fill_interleaved_mask_vec(sample_include, raw_sample_ctv, self._subset_include_interleaved_vec)
        
        cdef uint32_t raw_sample_ctl = DIV_UP(raw_sample_ct, kBitsPerWord)
        fill_cumulative_popcounts(sample_include, raw_sample_ctl, self._subset_cumulative_popcounts)

        self._subset_size = subset_size
        return

    
    def __cinit__(self, bytes filename, object raw_sample_ct = None,
                  object variant_ct = None, object sample_subset = None):
        self._info_ptr = <pgen_file_info_t*>PyMem_Malloc(sizeof(pgen_file_info_t))
        if not self._info_ptr:
            raise MemoryError()
        pgfi_preinit(self._info_ptr)
        # this depends on pgenlib_internal implementation.  could save
        # pgfi_alloc and pgr_alloc instead.
        self._info_ptr[0].vrtypes = NULL
        cdef uint32_t cur_sample_ct = 0xffffffffU
        if raw_sample_ct is not None:
            cur_sample_ct = raw_sample_ct
        cdef uint32_t cur_variant_ct = 0xffffffffU
        if variant_ct is not None:
            cur_variant_ct = variant_ct
        cdef const char* fname = <const char*>filename
        cdef pgen_header_ctrl_t header_ctrl
        cdef uintptr_t pgfi_alloc_cacheline_ct
        cdef char errstr_buf[kPglErrstrBufBlen]
        if pgfi_init_phase1(fname, cur_variant_ct, cur_sample_ct, 0, &header_ctrl, self._info_ptr, &pgfi_alloc_cacheline_ct, errstr_buf) != kPglRetSuccess:
            raise RuntimeError(errstr_buf[7:])
        assert (header_ctrl & 0x30) == 0 # no alt allele counts
        assert (header_ctrl & 0xc0) != 0xc0 # no explicit nonref_flags
        cdef uint32_t file_sample_ct = self._info_ptr[0].raw_sample_ct
        assert file_sample_ct != 0
        cdef unsigned char* pgfi_alloc = NULL
        if pgfi_alloc_cacheline_ct != 0:
            if cachealigned_malloc(pgfi_alloc_cacheline_ct * kCacheline, &pgfi_alloc):
                raise MemoryError()
        cdef uint32_t max_vrec_width
        cdef uintptr_t pgr_alloc_cacheline_ct
        if pgfi_init_phase2(header_ctrl, 1, 1, 0, 0, self._info_ptr[0].raw_variant_ct, &max_vrec_width, self._info_ptr, pgfi_alloc, &pgr_alloc_cacheline_ct, errstr_buf):
            if pgfi_alloc and not self._info_ptr[0].vrtypes:
                aligned_free(pgfi_alloc)
            raise RuntimeError(errstr_buf[7:])

        self._state_ptr = <pgen_reader_t*>PyMem_Malloc(sizeof(pgen_reader_t))
        if not self._state_ptr:
            raise MemoryError()
        pgr_preinit(self._state_ptr)
        self._state_ptr[0].fread_buf = NULL
        cdef uintptr_t pgr_alloc_main_byte_ct = pgr_alloc_cacheline_ct * kCacheline
        cdef uintptr_t sample_subset_byte_ct = DIV_UP(file_sample_ct, kBitsPerVec) * kBytesPerVec
        cdef uintptr_t cumulative_popcounts_byte_ct = DIV_UP(file_sample_ct, kBitsPerWord * kInt32PerVec) * kBytesPerVec
        cdef uintptr_t genovec_byte_ct = DIV_UP(file_sample_ct, kQuatersPerVec) * kBytesPerVec
        cdef uintptr_t dosage_vals_byte_ct = DIV_UP(file_sample_ct, (2 * kInt32PerVec)) * kBytesPerVec
        cdef unsigned char* pgr_alloc
        if cachealigned_malloc(pgr_alloc_main_byte_ct + (2 * kPglQuaterTransposeBatch + 5) * sample_subset_byte_ct + cumulative_popcounts_byte_ct + (1 + kPglQuaterTransposeBatch) * genovec_byte_ct + dosage_vals_byte_ct + kPglBitTransposeBufbytes + 4 * (kPglQuaterTransposeBatch * kPglQuaterTransposeBatch / 8), &pgr_alloc):
            raise MemoryError()
        cdef pglerr_t reterr = pgr_init(fname, max_vrec_width, self._info_ptr, self._state_ptr, pgr_alloc)
        if reterr != kPglRetSuccess:
            if not self._state_ptr[0].fread_buf:
                aligned_free(pgr_alloc)
            raise RuntimeError("pgl_init() error " + str(reterr))
        cdef unsigned char* pgr_alloc_iter = &(pgr_alloc[pgr_alloc_main_byte_ct])
        self._subset_include_vec = <uintptr_t*>pgr_alloc_iter
        pgr_alloc_iter = &(pgr_alloc_iter[sample_subset_byte_ct])
        self._subset_include_interleaved_vec = <uintptr_t*>pgr_alloc_iter
        pgr_alloc_iter = &(pgr_alloc_iter[sample_subset_byte_ct])
        
        # assumes kWordsPerVec <= 2
        self._subset_include_interleaved_vec[-1] = 0

        self._subset_cumulative_popcounts = <uint32_t*>pgr_alloc_iter
        pgr_alloc_iter = &(pgr_alloc_iter[cumulative_popcounts_byte_ct])
        self._genovec = <uintptr_t*>pgr_alloc_iter
        pgr_alloc_iter = &(pgr_alloc_iter[genovec_byte_ct])
        self._phasepresent = <uintptr_t*>pgr_alloc_iter
        pgr_alloc_iter = &(pgr_alloc_iter[sample_subset_byte_ct])
        self._phaseinfo = <uintptr_t*>pgr_alloc_iter
        pgr_alloc_iter = &(pgr_alloc_iter[sample_subset_byte_ct])
        self._dosage_present = <uintptr_t*>pgr_alloc_iter
        pgr_alloc_iter = &(pgr_alloc_iter[sample_subset_byte_ct])
        self._dosage_vals = <uint16_t*>pgr_alloc_iter
        pgr_alloc_iter = &(pgr_alloc_iter[dosage_vals_byte_ct])
        if sample_subset is not None:
            self.set_sample_subset_internal(sample_subset)
        else:
            self._subset_size = file_sample_ct
        self._transpose_batch_buf = <vul_t*>pgr_alloc_iter
        pgr_alloc_iter = &(pgr_alloc_iter[kPglBitTransposeBufbytes])
        self._multivar_vmaj_geno_buf = <uintptr_t*>pgr_alloc_iter
        pgr_alloc_iter = &(pgr_alloc_iter[kPglQuaterTransposeBatch * genovec_byte_ct])
        self._multivar_vmaj_phasepresent_buf = <uintptr_t*>pgr_alloc_iter
        pgr_alloc_iter = &(pgr_alloc_iter[kPglQuaterTransposeBatch * sample_subset_byte_ct])
        self._multivar_vmaj_phaseinfo_buf = <uintptr_t*>pgr_alloc_iter
        pgr_alloc_iter = &(pgr_alloc_iter[kPglQuaterTransposeBatch * sample_subset_byte_ct])
        self._multivar_smaj_geno_batch_buf = <uintptr_t*>pgr_alloc_iter
        pgr_alloc_iter = &(pgr_alloc_iter[kPglQuaterTransposeBatch * kPglQuaterTransposeBatch / 4])
        self._multivar_smaj_phaseinfo_batch_buf = <uintptr_t*>pgr_alloc_iter
        pgr_alloc_iter = &(pgr_alloc_iter[kPglQuaterTransposeBatch * kPglQuaterTransposeBatch / 8])
        self._multivar_smaj_phasepresent_batch_buf = <uintptr_t*>pgr_alloc_iter
        # pgr_alloc_iter = &(pgr_alloc_iter[kPglQuaterTransposeBatch * kPglQuaterTransposeBatch / 8])
        return

    
    cpdef __enter__(self):
        return self


    cpdef get_raw_sample_ct(self):
        return self._info_ptr[0].raw_sample_ct


    cpdef get_variant_ct(self):
        return self._info_ptr[0].raw_variant_ct


    cpdef hardcall_phase_present(self):
        return ((self._info_ptr[0].gflags & kfPgenGlobalHardcallPhasePresent) != 0)
    
    
    cpdef read(self, uint32_t variant_idx, np.ndarray geno_int_out, uint32_t allele_idx = 1):
        if variant_idx >= self._info_ptr[0].raw_variant_ct:
            # could have an unsafe mode which doesn't perform this check, but
            # let's default to at least this much bounds-checking
            raise RuntimeError("read() variant_idx too large (" + str(variant_idx) + "; only " + str(self._info_ptr[0].raw_variant_ct) + " in file)")
        if not geno_int_out.flags["C_CONTIGUOUS"]:
            raise RuntimeError("read() requires geno_int_out to be C-contiguous.")
        # for full genotype info for multiallelic variants, use read_phased()
        # instead
        cdef pglerr_t reterr = pgr_read_allele_countvec_subset_unsafe(self._subset_include_vec, self._subset_cumulative_popcounts, self._subset_size, variant_idx, allele_idx, self._state_ptr, self._genovec)
        if reterr != kPglRetSuccess:
            raise RuntimeError("read() error " + str(reterr))
        cdef int8_t* data8_ptr
        cdef int32_t* data32_ptr
        cdef int64_t* data64_ptr
        if geno_int_out.dtype == np.int8:
            data8_ptr = <int8_t*>geno_int_out.data
            genoarr_to_bytes_minus9(self._genovec, self._subset_size, data8_ptr)
        elif geno_int_out.dtype == np.int32:
            data32_ptr = <int32_t*>geno_int_out.data
            genoarr_to_int32s_minus9(self._genovec, self._subset_size, data32_ptr)
        elif geno_int_out.dtype == np.int64:
            data64_ptr = <int64_t*>geno_int_out.data
            genoarr_to_int64s_minus9(self._genovec, self._subset_size, data64_ptr)
        else:
            raise RuntimeError("Invalid read() geno_int_out array element type (int8, int32, or int64 expected).")
        return


    cpdef read_dosages(self, uint32_t variant_idx, np.ndarray floatarr_out, uint32_t allele_idx = 1):
        if variant_idx >= self._info_ptr[0].raw_variant_ct:
            raise RuntimeError("read_dosages() variant_idx too large (" + str(variant_idx) + "; only " + str(self._info_ptr[0].raw_variant_ct) + " in file)")
        if not floatarr_out.flags["C_CONTIGUOUS"]:
            raise RuntimeError("read_dosages() requires floatarr_out to be C-contiguous.")
        # todo: change this when pgenlib_internal supports multiallelic
        # variants
        cdef uint32_t dosage_ct
        cdef uint32_t is_explicit_alt1
        cdef pglerr_t reterr = pgr_read_refalt1_genovec_dosage16_subset_unsafe(self._subset_include_vec, self._subset_cumulative_popcounts, self._subset_size, variant_idx, self._state_ptr, self._genovec, self._dosage_present, self._dosage_vals, &dosage_ct, &is_explicit_alt1)
        if reterr != kPglRetSuccess:
            raise RuntimeError("read_dosages() error " + str(reterr))
        if allele_idx == 0:
            genovec_invert_unsafe(self._subset_size, self._genovec)
            biallelic_dosage16_invert(dosage_ct, self._dosage_vals)
        # todo: flip on allele_idx == 0
        cdef float* data32_ptr
        cdef double* data64_ptr
        if floatarr_out.dtype == np.float32:
            data32_ptr = <float*>floatarr_out.data
            dosage16_to_floats_minus9(self._genovec, self._dosage_present, self._dosage_vals, self._subset_size, dosage_ct, data32_ptr)
        elif floatarr_out.dtype == np.float64:
            data64_ptr = <double*>floatarr_out.data
            dosage16_to_doubles_minus9(self._genovec, self._dosage_present, self._dosage_vals, self._subset_size, dosage_ct, data64_ptr)
        else:
            raise RuntimeError("Invalid read_dosages() floatarr_out array element type (float32 or float64 expected).")
        return


    cpdef read_alleles(self, uint32_t variant_idx, np.ndarray[np.int32_t,mode="c",ndim=1] allele_int32_out):
        if variant_idx >= self._info_ptr[0].raw_variant_ct:
            # could have an unsafe mode which doesn't perform this check, but
            # let's default to at least this much bounds-checking
            raise RuntimeError("read_alleles() variant_idx too large (" + str(variant_idx) + "; only " + str(self._info_ptr[0].raw_variant_ct) + " in file)")
        cdef uint32_t phasepresent_ct
        # upgrade to multiallelic version of this function in the future
        cdef pglerr_t reterr = pgr_read_refalt1_genovec_hphase_subset_unsafe(self._subset_include_vec, self._subset_cumulative_popcounts, self._subset_size, variant_idx, self._state_ptr, self._genovec, self._phasepresent, self._phaseinfo, &phasepresent_ct)
        if reterr != kPglRetSuccess:
            raise RuntimeError("read_alleles() error " + str(reterr))
        cdef int32_t* main_data_ptr = <int32_t*>(&(allele_int32_out[0]))
        genoarr_phased_to_allele_codes(self._genovec, self._phasepresent, self._phaseinfo, self._subset_size, phasepresent_ct, NULL, main_data_ptr)
        return


    cpdef read_alleles_and_phasepresent(self, uint32_t variant_idx, np.ndarray[np.int32_t,mode="c",ndim=1] allele_int32_out, np.ndarray[np.uint8_t,mode="c",cast=True] phasepresent_out):
        if variant_idx >= self._info_ptr[0].raw_variant_ct:
            # could have an unsafe mode which doesn't perform this check, but
            # let's default to at least this much bounds-checking
            raise RuntimeError("read_alleles_and_phasepresent() variant_idx too large (" + str(variant_idx) + "; only " + str(self._info_ptr[0].raw_variant_ct) + " in file)")
        cdef uint32_t phasepresent_ct
        # upgrade to multiallelic version of this function in the future
        cdef pglerr_t reterr = pgr_read_refalt1_genovec_hphase_subset_unsafe(self._subset_include_vec, self._subset_cumulative_popcounts, self._subset_size, variant_idx, self._state_ptr, self._genovec, self._phasepresent, self._phaseinfo, &phasepresent_ct)
        if reterr != kPglRetSuccess:
            raise RuntimeError("read_alleles_and_phasepresent() error " + str(reterr))
        cdef int32_t* main_data_ptr = <int32_t*>(&(allele_int32_out[0]))
        cdef unsigned char* phasepresent_data_ptr = <unsigned char*>(&(phasepresent_out[0]))
        genoarr_phased_to_allele_codes(self._genovec, self._phasepresent, self._phaseinfo, self._subset_size, phasepresent_ct, phasepresent_data_ptr, main_data_ptr)
        return


    cdef read_range_internal8(self, uint32_t variant_idx_start, uint32_t variant_idx_end, np.ndarray[np.int8_t,mode="c",ndim=2] geno_int8_out, uint32_t allele_idx = 1, bint sample_maj = 0):
        if variant_idx_end > self._info_ptr[0].raw_variant_ct:
            raise RuntimeError("read_range() variant_idx_end too large (" + str(variant_idx_end) + "; only " + str(self._info_ptr[0].raw_variant_ct) + " in file)")
        cdef const uintptr_t* subset_include_vec = self._subset_include_vec
        cdef const uint32_t* subset_cumulative_popcounts = self._subset_cumulative_popcounts
        cdef pgen_reader_t* pgrp = self._state_ptr
        cdef uintptr_t* genovec = self._genovec
        cdef uint32_t variant_idx_ct = variant_idx_end - variant_idx_start
        cdef uint32_t subset_size = self._subset_size
        cdef int8_t* data_ptr
        cdef uint32_t variant_idx
        cdef pglerr_t reterr
        if sample_maj == 0:
            if geno_int8_out.shape[0] < variant_idx_ct:
                raise RuntimeError("Variant-major read_range() geno_int_out buffer has too few rows (" + str(geno_int8_out.shape[0]) + "; (variant_idx_end - variant_idx_start) is " + str(variant_idx_ct) + ")")
            if geno_int8_out.shape[1] < subset_size:
                raise RuntimeError("Variant-major read_range() geno_int_out buffer has too few columns (" + str(geno_int8_out.shape[1]) + "; current sample subset has size " + str(subset_size) + ")")
            for variant_idx in range(variant_idx_start, variant_idx_end):
                reterr = pgr_read_allele_countvec_subset_unsafe(subset_include_vec, subset_cumulative_popcounts, subset_size, variant_idx, allele_idx, pgrp, genovec)
                if reterr != kPglRetSuccess:
                    raise RuntimeError("read_range() error " + str(reterr))
                data_ptr = &(geno_int8_out[(variant_idx - variant_idx_start), 0])
                genoarr_to_bytes_minus9(genovec, subset_size, data_ptr)
            return
        if variant_idx_start >= variant_idx_end:
            raise RuntimeError("read_range() variant_idx_start >= variant_idx_end (" + str(variant_idx_start) + ", " + str(variant_idx_end) + ")")
        if geno_int8_out.shape[0] < subset_size:
            raise RuntimeError("Sample-major read_range() geno_int_out buffer has too few rows (" + str(geno_int8_out.shape[0]) + "; current sample subset has size " + str(subset_size) + ")")
        if geno_int8_out.shape[1] < variant_idx_ct:
            raise RuntimeError("Sample-major read_range() geno_int_out buffer has too few columns (" + str(geno_int8_out.shape[1]) + "; (variant_idx_end - variant_idx_start) is " + str(variant_idx_ct) + ")")
        cdef uint32_t variant_batch_ct = DIV_UP(variant_idx_ct, kPglQuaterTransposeBatch)
        cdef uint32_t variant_batch_size = kPglQuaterTransposeBatch
        cdef uint32_t variant_idx_offset = variant_idx_start
        cdef uint32_t sample_ctaw2 = kWordsPerVec * DIV_UP(subset_size, kBitsPerWordD2)
        cdef uint32_t sample_batch_ct = DIV_UP(subset_size, kPglQuaterTransposeBatch)
        cdef vul_t* transpose_batch_buf = self._transpose_batch_buf
        cdef uintptr_t* multivar_vmaj_geno_buf = self._multivar_vmaj_geno_buf
        cdef uintptr_t* multivar_smaj_geno_batch_buf = self._multivar_smaj_geno_batch_buf
        cdef uintptr_t* vmaj_iter
        cdef uintptr_t* smaj_iter
        cdef uint32_t variant_batch_idx
        cdef uint32_t sample_batch_size
        cdef uint32_t sample_batch_idx
        cdef uint32_t uii
        for variant_batch_idx in range(variant_batch_ct):
            if variant_batch_idx == (variant_batch_ct - 1):
                variant_batch_size = 1 + <uint32_t>((variant_idx_ct - 1) % kPglQuaterTransposeBatch)
            vmaj_iter = multivar_vmaj_geno_buf
            for uii in range(variant_batch_size):
                reterr = pgr_read_allele_countvec_subset_unsafe(subset_include_vec, subset_cumulative_popcounts, subset_size, uii + variant_idx_offset, allele_idx, pgrp, vmaj_iter)
                if reterr != kPglRetSuccess:
                    raise RuntimeError("read_range() error " + str(reterr))
                vmaj_iter = &(vmaj_iter[sample_ctaw2])
            sample_batch_size = kPglQuaterTransposeBatch
            vmaj_iter = multivar_vmaj_geno_buf
            for sample_batch_idx in range(sample_batch_ct):
                if sample_batch_idx == sample_batch_ct - 1:
                    sample_batch_size = 1 + <uint32_t>((subset_size - 1) % kPglQuaterTransposeBatch)
                smaj_iter = multivar_smaj_geno_batch_buf
                transpose_quaterblock(vmaj_iter, sample_ctaw2, kPglQuaterTransposeWords, variant_batch_size, sample_batch_size, smaj_iter, transpose_batch_buf)
                for uii in range(sample_batch_size):
                    data_ptr = &(geno_int8_out[uii + sample_batch_idx * kPglQuaterTransposeBatch, variant_batch_idx * kPglQuaterTransposeBatch])
                    genoarr_to_bytes_minus9(smaj_iter, variant_batch_size, data_ptr)
                    smaj_iter = &(smaj_iter[kPglQuaterTransposeWords])
                vmaj_iter = &(vmaj_iter[kPglQuaterTransposeWords])
            variant_idx_offset += kPglQuaterTransposeBatch
        return

    cdef read_range_internal32(self, uint32_t variant_idx_start, uint32_t variant_idx_end, np.ndarray[np.int32_t,mode="c",ndim=2] geno_int32_out, uint32_t allele_idx = 1, bint sample_maj = 0):
        if variant_idx_end > self._info_ptr[0].raw_variant_ct:
            raise RuntimeError("read_range() variant_idx_end too large (" + str(variant_idx_end) + "; only " + str(self._info_ptr[0].raw_variant_ct) + " in file)")
        cdef const uintptr_t* subset_include_vec = self._subset_include_vec
        cdef const uint32_t* subset_cumulative_popcounts = self._subset_cumulative_popcounts
        cdef pgen_reader_t* pgrp = self._state_ptr
        cdef uintptr_t* genovec = self._genovec
        cdef uint32_t variant_idx_ct = variant_idx_end - variant_idx_start
        cdef uint32_t subset_size = self._subset_size
        cdef int32_t* data_ptr
        cdef uint32_t variant_idx
        cdef pglerr_t reterr
        if sample_maj == 0:
            if geno_int32_out.shape[0] < variant_idx_ct:
                raise RuntimeError("Variant-major read_range() geno_int_out buffer has too few rows (" + str(geno_int32_out.shape[0]) + "; (variant_idx_end - variant_idx_start) is " + str(variant_idx_ct) + ")")
            if geno_int32_out.shape[1] < subset_size:
                raise RuntimeError("Variant-major read_range() geno_int_out buffer has too few columns (" + str(geno_int32_out.shape[1]) + "; current sample subset has size " + str(subset_size) + ")")
            for variant_idx in range(variant_idx_start, variant_idx_end):
                reterr = pgr_read_allele_countvec_subset_unsafe(subset_include_vec, subset_cumulative_popcounts, subset_size, variant_idx, allele_idx, pgrp, genovec)
                if reterr != kPglRetSuccess:
                    raise RuntimeError("read_range() error " + str(reterr))
                data_ptr = <int32_t*>(&(geno_int32_out[(variant_idx - variant_idx_start), 0]))
                genoarr_to_int32s_minus9(genovec, subset_size, data_ptr)
            return
        if variant_idx_start >= variant_idx_end:
            raise RuntimeError("read_range() variant_idx_start >= variant_idx_end (" + str(variant_idx_start) + ", " + str(variant_idx_end) + ")")
        if geno_int32_out.shape[0] < subset_size:
            raise RuntimeError("Sample-major read_range() geno_int_out buffer has too few rows (" + str(geno_int32_out.shape[0]) + "; current sample subset has size " + str(subset_size) + ")")
        if geno_int32_out.shape[1] < variant_idx_ct:
            raise RuntimeError("Sample-major read_range() geno_int_out buffer has too few columns (" + str(geno_int32_out.shape[1]) + "; (variant_idx_end - variant_idx_start) is " + str(variant_idx_ct) + ")")
        cdef uint32_t variant_batch_ct = DIV_UP(variant_idx_ct, kPglQuaterTransposeBatch)
        cdef uint32_t variant_batch_size = kPglQuaterTransposeBatch
        cdef uint32_t variant_idx_offset = variant_idx_start
        cdef uint32_t sample_ctaw2 = kWordsPerVec * DIV_UP(subset_size, kBitsPerWordD2)
        cdef uint32_t sample_batch_ct = DIV_UP(subset_size, kPglQuaterTransposeBatch)
        cdef vul_t* transpose_batch_buf = self._transpose_batch_buf
        cdef uintptr_t* multivar_vmaj_geno_buf = self._multivar_vmaj_geno_buf
        cdef uintptr_t* multivar_smaj_geno_batch_buf = self._multivar_smaj_geno_batch_buf
        cdef uintptr_t* vmaj_iter
        cdef uintptr_t* smaj_iter
        cdef uint32_t variant_batch_idx
        cdef uint32_t sample_batch_size
        cdef uint32_t sample_batch_idx
        cdef uint32_t uii
        for variant_batch_idx in range(variant_batch_ct):
            if variant_batch_idx == (variant_batch_ct - 1):
                variant_batch_size = 1 + <uint32_t>((variant_idx_ct - 1) % kPglQuaterTransposeBatch)
            vmaj_iter = multivar_vmaj_geno_buf
            for uii in range(variant_batch_size):
                reterr = pgr_read_allele_countvec_subset_unsafe(subset_include_vec, subset_cumulative_popcounts, subset_size, uii + variant_idx_offset, allele_idx, pgrp, vmaj_iter)
                if reterr != kPglRetSuccess:
                    raise RuntimeError("read_range() error " + str(reterr))
                vmaj_iter = &(vmaj_iter[sample_ctaw2])
            sample_batch_size = kPglQuaterTransposeBatch
            vmaj_iter = multivar_vmaj_geno_buf
            for sample_batch_idx in range(sample_batch_ct):
                if sample_batch_idx == sample_batch_ct - 1:
                    sample_batch_size = 1 + <uint32_t>((subset_size - 1) % kPglQuaterTransposeBatch)
                smaj_iter = multivar_smaj_geno_batch_buf
                transpose_quaterblock(vmaj_iter, sample_ctaw2, kPglQuaterTransposeWords, variant_batch_size, sample_batch_size, smaj_iter, transpose_batch_buf)
                for uii in range(sample_batch_size):
                    data_ptr = <int32_t*>(&(geno_int32_out[uii + sample_batch_idx * kPglQuaterTransposeBatch, variant_batch_idx * kPglQuaterTransposeBatch]))
                    genoarr_to_int32s_minus9(smaj_iter, variant_batch_size, data_ptr)
                    smaj_iter = &(smaj_iter[kPglQuaterTransposeWords])
                vmaj_iter = &(vmaj_iter[kPglQuaterTransposeWords])
            variant_idx_offset += kPglQuaterTransposeBatch
        return

    cdef read_range_internal64(self, uint32_t variant_idx_start, uint32_t variant_idx_end, np.ndarray[np.int64_t,mode="c",ndim=2] geno_int64_out, uint32_t allele_idx = 1, bint sample_maj = 0):
        if variant_idx_end > self._info_ptr[0].raw_variant_ct:
            raise RuntimeError("read_range() variant_idx_end too large (" + str(variant_idx_end) + "; only " + str(self._info_ptr[0].raw_variant_ct) + " in file)")
        cdef const uintptr_t* subset_include_vec = self._subset_include_vec
        cdef const uint32_t* subset_cumulative_popcounts = self._subset_cumulative_popcounts
        cdef pgen_reader_t* pgrp = self._state_ptr
        cdef uintptr_t* genovec = self._genovec
        cdef uint32_t variant_idx_ct = variant_idx_end - variant_idx_start
        cdef uint32_t subset_size = self._subset_size
        cdef int64_t* data_ptr
        cdef uint32_t variant_idx
        cdef pglerr_t reterr
        if sample_maj == 0:
            if geno_int64_out.shape[0] < variant_idx_ct:
                raise RuntimeError("Variant-major read_range() geno_int_out buffer has too few rows (" + str(geno_int64_out.shape[0]) + "; (variant_idx_end - variant_idx_start) is " + str(variant_idx_ct) + ")")
            if geno_int64_out.shape[1] < subset_size:
                raise RuntimeError("Variant-major read_range() geno_int_out buffer has too few columns (" + str(geno_int64_out.shape[1]) + "; current sample subset has size " + str(subset_size) + ")")
            for variant_idx in range(variant_idx_start, variant_idx_end):
                reterr = pgr_read_allele_countvec_subset_unsafe(subset_include_vec, subset_cumulative_popcounts, subset_size, variant_idx, allele_idx, pgrp, genovec)
                if reterr != kPglRetSuccess:
                    raise RuntimeError("read_range() error " + str(reterr))
                data_ptr = &(geno_int64_out[(variant_idx - variant_idx_start), 0])
                genoarr_to_int64s_minus9(genovec, subset_size, data_ptr)
            return
        if variant_idx_start >= variant_idx_end:
            raise RuntimeError("read_range() variant_idx_start >= variant_idx_end (" + str(variant_idx_start) + ", " + str(variant_idx_end) + ")")
        if geno_int64_out.shape[0] < subset_size:
            raise RuntimeError("Sample-major read_range() geno_int_out buffer has too few rows (" + str(geno_int64_out.shape[0]) + "; current sample subset has size " + str(subset_size) + ")")
        if geno_int64_out.shape[1] < variant_idx_ct:
            raise RuntimeError("Sample-major read_range() geno_int_out buffer has too few columns (" + str(geno_int64_out.shape[1]) + "; (variant_idx_end - variant_idx_start) is " + str(variant_idx_ct) + ")")
        cdef uint32_t variant_batch_ct = DIV_UP(variant_idx_ct, kPglQuaterTransposeBatch)
        cdef uint32_t variant_batch_size = kPglQuaterTransposeBatch
        cdef uint32_t variant_idx_offset = variant_idx_start
        cdef uint32_t sample_ctaw2 = kWordsPerVec * DIV_UP(subset_size, kBitsPerWordD2)
        cdef uint32_t sample_batch_ct = DIV_UP(subset_size, kPglQuaterTransposeBatch)
        cdef vul_t* transpose_batch_buf = self._transpose_batch_buf
        cdef uintptr_t* multivar_vmaj_geno_buf = self._multivar_vmaj_geno_buf
        cdef uintptr_t* multivar_smaj_geno_batch_buf = self._multivar_smaj_geno_batch_buf
        cdef uintptr_t* vmaj_iter
        cdef uintptr_t* smaj_iter
        cdef uint32_t variant_batch_idx
        cdef uint32_t sample_batch_size
        cdef uint32_t sample_batch_idx
        cdef uint32_t uii
        for variant_batch_idx in range(variant_batch_ct):
            if variant_batch_idx == (variant_batch_ct - 1):
                variant_batch_size = 1 + <uint32_t>((variant_idx_ct - 1) % kPglQuaterTransposeBatch)
            vmaj_iter = multivar_vmaj_geno_buf
            for uii in range(variant_batch_size):
                reterr = pgr_read_allele_countvec_subset_unsafe(subset_include_vec, subset_cumulative_popcounts, subset_size, uii + variant_idx_offset, allele_idx, pgrp, vmaj_iter)
                if reterr != kPglRetSuccess:
                    raise RuntimeError("read_range() error " + str(reterr))
                vmaj_iter = &(vmaj_iter[sample_ctaw2])
            sample_batch_size = kPglQuaterTransposeBatch
            vmaj_iter = multivar_vmaj_geno_buf
            for sample_batch_idx in range(sample_batch_ct):
                if sample_batch_idx == sample_batch_ct - 1:
                    sample_batch_size = 1 + <uint32_t>((subset_size - 1) % kPglQuaterTransposeBatch)
                smaj_iter = multivar_smaj_geno_batch_buf
                transpose_quaterblock(vmaj_iter, sample_ctaw2, kPglQuaterTransposeWords, variant_batch_size, sample_batch_size, smaj_iter, transpose_batch_buf)
                for uii in range(sample_batch_size):
                    data_ptr = &(geno_int64_out[uii + sample_batch_idx * kPglQuaterTransposeBatch, variant_batch_idx * kPglQuaterTransposeBatch])
                    genoarr_to_int64s_minus9(smaj_iter, variant_batch_size, data_ptr)
                    smaj_iter = &(smaj_iter[kPglQuaterTransposeWords])
                vmaj_iter = &(vmaj_iter[kPglQuaterTransposeWords])
            variant_idx_offset += kPglQuaterTransposeBatch
        return

    cpdef read_range(self, uint32_t variant_idx_start, uint32_t variant_idx_end, np.ndarray geno_int_out, uint32_t allele_idx = 1, bint sample_maj = 0):
        # C-contiguity checked by read_range_internal8(), etc.
        if geno_int_out.dtype == np.int8:
            self.read_range_internal8(variant_idx_start, variant_idx_end, geno_int_out, allele_idx, sample_maj)
        elif geno_int_out.dtype == np.int32:
            self.read_range_internal32(variant_idx_start, variant_idx_end, geno_int_out, allele_idx, sample_maj)
        elif geno_int_out.dtype == np.int64:
            self.read_range_internal64(variant_idx_start, variant_idx_end, geno_int_out, allele_idx, sample_maj)
        else:
            raise RuntimeError("Invalid read_range() geno_int_out array element type (int8, int32, or int64 expected).")
        return


    cdef read_list_internal8(self, np.ndarray[np.uint32_t] variant_idxs, np.ndarray[np.int8_t,mode="c",ndim=2] geno_int8_out, uint32_t allele_idx = 1, bint sample_maj = 0):
        cdef uint32_t raw_variant_ct = self._info_ptr[0].raw_variant_ct
        cdef const uintptr_t* subset_include_vec = self._subset_include_vec
        cdef const uint32_t* subset_cumulative_popcounts = self._subset_cumulative_popcounts
        cdef pgen_reader_t* pgrp = self._state_ptr
        cdef uintptr_t* genovec = self._genovec
        cdef uint32_t variant_idx_ct = <uint32_t>variant_idxs.shape[0]
        cdef uint32_t subset_size = self._subset_size
        cdef int8_t* data_ptr
        cdef uint32_t variant_list_idx
        cdef uint32_t variant_idx
        cdef pglerr_t reterr
        if sample_maj == 0:
            if geno_int8_out.shape[0] < variant_idx_ct:
                raise RuntimeError("Variant-major read_list() geno_int_out buffer has too few rows (" + str(geno_int8_out.shape[0]) + "; variant_idxs length is " + str(variant_idx_ct) + ")")
            if geno_int8_out.shape[1] < subset_size:
                raise RuntimeError("Variant-major read_list() geno_int_out buffer has too few columns (" + str(geno_int8_out.shape[1]) + "; current sample subset has size " + str(subset_size) + ")")
            for variant_list_idx in range(variant_idx_ct):
                variant_idx = variant_idxs[variant_list_idx]
                if variant_idx >= raw_variant_ct:
                    raise RuntimeError("read_list() variant index too large (" + str(variant_idx) + "; only " + str(raw_variant_ct) + " in file)")
                reterr = pgr_read_allele_countvec_subset_unsafe(subset_include_vec, subset_cumulative_popcounts, subset_size, variant_idx, allele_idx, pgrp, genovec)
                if reterr != kPglRetSuccess:
                    raise RuntimeError("read_range() error " + str(reterr))
                data_ptr = &(geno_int8_out[variant_list_idx, 0])
                genoarr_to_bytes_minus9(genovec, subset_size, data_ptr)
            return
        if geno_int8_out.shape[0] < subset_size:
            raise RuntimeError("Sample-major read_list() geno_int_out buffer has too few rows (" + str(geno_int8_out.shape[0]) + "; current sample subset has size " + str(subset_size) + ")")
        if geno_int8_out.shape[1] < variant_idx_ct:
            raise RuntimeError("Sample-major read_list() geno_int_out buffer has too few columns (" + str(geno_int8_out.shape[1]) + "; variant_idxs length is " + str(variant_idx_ct) + ")")
        cdef uint32_t variant_batch_ct = DIV_UP(variant_idx_ct, kPglQuaterTransposeBatch)
        cdef uint32_t variant_batch_size = kPglQuaterTransposeBatch
        cdef uint32_t sample_ctaw2 = kWordsPerVec * DIV_UP(subset_size, kBitsPerWordD2)
        cdef uint32_t sample_batch_ct = DIV_UP(subset_size, kPglQuaterTransposeBatch)
        cdef vul_t* transpose_batch_buf = self._transpose_batch_buf
        cdef uintptr_t* multivar_vmaj_geno_buf = self._multivar_vmaj_geno_buf
        cdef uintptr_t* multivar_smaj_geno_batch_buf = self._multivar_smaj_geno_batch_buf
        cdef uintptr_t* vmaj_iter
        cdef uintptr_t* smaj_iter
        cdef uint32_t variant_batch_idx
        cdef uint32_t sample_batch_size
        cdef uint32_t sample_batch_idx
        cdef uint32_t uii
        variant_list_idx = 0
        for variant_batch_idx in range(variant_batch_ct):
            if variant_batch_idx == (variant_batch_ct - 1):
                variant_batch_size = 1 + <uint32_t>((variant_idx_ct - 1) % kPglQuaterTransposeBatch)
            vmaj_iter = multivar_vmaj_geno_buf
            for uii in range(variant_batch_size):
                variant_idx = variant_idxs[uii + variant_list_idx]
                if variant_idx >= raw_variant_ct:
                    raise RuntimeError("read_list() variant index too large (" + str(variant_idx) + "; only " + str(raw_variant_ct) + " in file)")
                reterr = pgr_read_allele_countvec_subset_unsafe(subset_include_vec, subset_cumulative_popcounts, subset_size, variant_idx, allele_idx, pgrp, vmaj_iter)
                if reterr != kPglRetSuccess:
                    raise RuntimeError("read_list() error " + str(reterr))
                vmaj_iter = &(vmaj_iter[sample_ctaw2])
            sample_batch_size = kPglQuaterTransposeBatch
            vmaj_iter = multivar_vmaj_geno_buf
            for sample_batch_idx in range(sample_batch_ct):
                if sample_batch_idx == sample_batch_ct - 1:
                    sample_batch_size = 1 + <uint32_t>((subset_size - 1) % kPglQuaterTransposeBatch)
                smaj_iter = multivar_smaj_geno_batch_buf
                transpose_quaterblock(vmaj_iter, sample_ctaw2, kPglQuaterTransposeWords, variant_batch_size, sample_batch_size, smaj_iter, transpose_batch_buf)
                for uii in range(sample_batch_size):
                    data_ptr = &(geno_int8_out[uii + sample_batch_idx * kPglQuaterTransposeBatch, variant_batch_idx * kPglQuaterTransposeBatch])
                    genoarr_to_bytes_minus9(smaj_iter, variant_batch_size, data_ptr)
                    smaj_iter = &(smaj_iter[kPglQuaterTransposeWords])
                vmaj_iter = &(vmaj_iter[kPglQuaterTransposeWords])
            variant_list_idx += kPglQuaterTransposeBatch
        return

    cdef read_list_internal32(self, np.ndarray[np.uint32_t] variant_idxs, np.ndarray[np.int32_t,mode="c",ndim=2] geno_int32_out, uint32_t allele_idx = 1, bint sample_maj = 0):
        cdef uint32_t raw_variant_ct = self._info_ptr[0].raw_variant_ct
        cdef const uintptr_t* subset_include_vec = self._subset_include_vec
        cdef const uint32_t* subset_cumulative_popcounts = self._subset_cumulative_popcounts
        cdef pgen_reader_t* pgrp = self._state_ptr
        cdef uintptr_t* genovec = self._genovec
        cdef uint32_t variant_idx_ct = <uint32_t>variant_idxs.shape[0]
        cdef uint32_t subset_size = self._subset_size
        cdef int32_t* data_ptr
        cdef uint32_t variant_list_idx
        cdef uint32_t variant_idx
        cdef pglerr_t reterr
        if sample_maj == 0:
            if geno_int32_out.shape[0] < variant_idx_ct:
                raise RuntimeError("Variant-major read_list() geno_int_out buffer has too few rows (" + str(geno_int32_out.shape[0]) + "; variant_idxs length is " + str(variant_idx_ct) + ")")
            if geno_int32_out.shape[1] < subset_size:
                raise RuntimeError("Variant-major read_list() geno_int_out buffer has too few columns (" + str(geno_int32_out.shape[1]) + "; current sample subset has size " + str(subset_size) + ")")
            for variant_list_idx in range(variant_idx_ct):
                variant_idx = variant_idxs[variant_list_idx]
                if variant_idx >= raw_variant_ct:
                    raise RuntimeError("read_list() variant index too large (" + str(variant_idx) + "; only " + str(raw_variant_ct) + " in file)")
                reterr = pgr_read_allele_countvec_subset_unsafe(subset_include_vec, subset_cumulative_popcounts, subset_size, variant_idx, allele_idx, pgrp, genovec)
                if reterr != kPglRetSuccess:
                    raise RuntimeError("read_range() error " + str(reterr))
                data_ptr = <int32_t*>(&(geno_int32_out[variant_list_idx, 0]))
                genoarr_to_int32s_minus9(genovec, subset_size, data_ptr)
            return
        if geno_int32_out.shape[0] < subset_size:
            raise RuntimeError("Sample-major read_list() geno_int_out buffer has too few rows (" + str(geno_int32_out.shape[0]) + "; current sample subset has size " + str(subset_size) + ")")
        if geno_int32_out.shape[1] < variant_idx_ct:
            raise RuntimeError("Sample-major read_list() geno_int_out buffer has too few columns (" + str(geno_int32_out.shape[1]) + "; variant_idxs length is " + str(variant_idx_ct) + ")")
        cdef uint32_t variant_batch_ct = DIV_UP(variant_idx_ct, kPglQuaterTransposeBatch)
        cdef uint32_t variant_batch_size = kPglQuaterTransposeBatch
        cdef uint32_t sample_ctaw2 = kWordsPerVec * DIV_UP(subset_size, kBitsPerWordD2)
        cdef uint32_t sample_batch_ct = DIV_UP(subset_size, kPglQuaterTransposeBatch)
        cdef vul_t* transpose_batch_buf = self._transpose_batch_buf
        cdef uintptr_t* multivar_vmaj_geno_buf = self._multivar_vmaj_geno_buf
        cdef uintptr_t* multivar_smaj_geno_batch_buf = self._multivar_smaj_geno_batch_buf
        cdef uintptr_t* vmaj_iter
        cdef uintptr_t* smaj_iter
        cdef uint32_t variant_batch_idx
        cdef uint32_t sample_batch_size
        cdef uint32_t sample_batch_idx
        cdef uint32_t uii
        variant_list_idx = 0
        for variant_batch_idx in range(variant_batch_ct):
            if variant_batch_idx == (variant_batch_ct - 1):
                variant_batch_size = 1 + <uint32_t>((variant_idx_ct - 1) % kPglQuaterTransposeBatch)
            vmaj_iter = multivar_vmaj_geno_buf
            for uii in range(variant_batch_size):
                variant_idx = variant_idxs[uii + variant_list_idx]
                if variant_idx >= raw_variant_ct:
                    raise RuntimeError("read_list() variant index too large (" + str(variant_idx) + "; only " + str(raw_variant_ct) + " in file)")
                reterr = pgr_read_allele_countvec_subset_unsafe(subset_include_vec, subset_cumulative_popcounts, subset_size, variant_idx, allele_idx, pgrp, vmaj_iter)
                if reterr != kPglRetSuccess:
                    raise RuntimeError("read_list() error " + str(reterr))
                vmaj_iter = &(vmaj_iter[sample_ctaw2])
            sample_batch_size = kPglQuaterTransposeBatch
            vmaj_iter = multivar_vmaj_geno_buf
            for sample_batch_idx in range(sample_batch_ct):
                if sample_batch_idx == sample_batch_ct - 1:
                    sample_batch_size = 1 + <uint32_t>((subset_size - 1) % kPglQuaterTransposeBatch)
                smaj_iter = multivar_smaj_geno_batch_buf
                transpose_quaterblock(vmaj_iter, sample_ctaw2, kPglQuaterTransposeWords, variant_batch_size, sample_batch_size, smaj_iter, transpose_batch_buf)
                for uii in range(sample_batch_size):
                    data_ptr = <int32_t*>(&(geno_int32_out[uii + sample_batch_idx * kPglQuaterTransposeBatch, variant_batch_idx * kPglQuaterTransposeBatch]))
                    genoarr_to_int32s_minus9(smaj_iter, variant_batch_size, data_ptr)
                    smaj_iter = &(smaj_iter[kPglQuaterTransposeWords])
                vmaj_iter = &(vmaj_iter[kPglQuaterTransposeWords])
            variant_list_idx += kPglQuaterTransposeBatch
        return

    cdef read_list_internal64(self, np.ndarray[np.uint32_t] variant_idxs, np.ndarray[np.int64_t,mode="c",ndim=2] geno_int64_out, uint32_t allele_idx = 1, bint sample_maj = 0):
        cdef uint32_t raw_variant_ct = self._info_ptr[0].raw_variant_ct
        cdef const uintptr_t* subset_include_vec = self._subset_include_vec
        cdef const uint32_t* subset_cumulative_popcounts = self._subset_cumulative_popcounts
        cdef pgen_reader_t* pgrp = self._state_ptr
        cdef uintptr_t* genovec = self._genovec
        cdef uint32_t variant_idx_ct = <uint32_t>variant_idxs.shape[0]
        cdef uint32_t subset_size = self._subset_size
        cdef int64_t* data_ptr
        cdef uint32_t variant_list_idx
        cdef uint32_t variant_idx
        cdef pglerr_t reterr
        if sample_maj == 0:
            if geno_int64_out.shape[0] < variant_idx_ct:
                raise RuntimeError("Variant-major read_list() geno_int_out buffer has too few rows (" + str(geno_int64_out.shape[0]) + "; variant_idxs length is " + str(variant_idx_ct) + ")")
            if geno_int64_out.shape[1] < subset_size:
                raise RuntimeError("Variant-major read_list() geno_int_out buffer has too few columns (" + str(geno_int64_out.shape[1]) + "; current sample subset has size " + str(subset_size) + ")")
            for variant_list_idx in range(variant_idx_ct):
                variant_idx = variant_idxs[variant_list_idx]
                if variant_idx >= raw_variant_ct:
                    raise RuntimeError("read_list() variant index too large (" + str(variant_idx) + "; only " + str(raw_variant_ct) + " in file)")
                reterr = pgr_read_allele_countvec_subset_unsafe(subset_include_vec, subset_cumulative_popcounts, subset_size, variant_idx, allele_idx, pgrp, genovec)
                if reterr != kPglRetSuccess:
                    raise RuntimeError("read_range() error " + str(reterr))
                data_ptr = &(geno_int64_out[variant_list_idx, 0])
                genoarr_to_int64s_minus9(genovec, subset_size, data_ptr)
            return
        if geno_int64_out.shape[0] < subset_size:
            raise RuntimeError("Sample-major read_list() geno_int_out buffer has too few rows (" + str(geno_int64_out.shape[0]) + "; current sample subset has size " + str(subset_size) + ")")
        if geno_int64_out.shape[1] < variant_idx_ct:
            raise RuntimeError("Sample-major read_list() geno_int_out buffer has too few columns (" + str(geno_int64_out.shape[1]) + "; variant_idxs length is " + str(variant_idx_ct) + ")")
        cdef uint32_t variant_batch_ct = DIV_UP(variant_idx_ct, kPglQuaterTransposeBatch)
        cdef uint32_t variant_batch_size = kPglQuaterTransposeBatch
        cdef uint32_t sample_ctaw2 = kWordsPerVec * DIV_UP(subset_size, kBitsPerWordD2)
        cdef uint32_t sample_batch_ct = DIV_UP(subset_size, kPglQuaterTransposeBatch)
        cdef vul_t* transpose_batch_buf = self._transpose_batch_buf
        cdef uintptr_t* multivar_vmaj_geno_buf = self._multivar_vmaj_geno_buf
        cdef uintptr_t* multivar_smaj_geno_batch_buf = self._multivar_smaj_geno_batch_buf
        cdef uintptr_t* vmaj_iter
        cdef uintptr_t* smaj_iter
        cdef uint32_t variant_batch_idx
        cdef uint32_t sample_batch_size
        cdef uint32_t sample_batch_idx
        cdef uint32_t uii
        variant_list_idx = 0
        for variant_batch_idx in range(variant_batch_ct):
            if variant_batch_idx == (variant_batch_ct - 1):
                variant_batch_size = 1 + <uint32_t>((variant_idx_ct - 1) % kPglQuaterTransposeBatch)
            vmaj_iter = multivar_vmaj_geno_buf
            for uii in range(variant_batch_size):
                variant_idx = variant_idxs[uii + variant_list_idx]
                if variant_idx >= raw_variant_ct:
                    raise RuntimeError("read_list() variant index too large (" + str(variant_idx) + "; only " + str(raw_variant_ct) + " in file)")
                reterr = pgr_read_allele_countvec_subset_unsafe(subset_include_vec, subset_cumulative_popcounts, subset_size, variant_idx, allele_idx, pgrp, vmaj_iter)
                if reterr != kPglRetSuccess:
                    raise RuntimeError("read_list() error " + str(reterr))
                vmaj_iter = &(vmaj_iter[sample_ctaw2])
            sample_batch_size = kPglQuaterTransposeBatch
            vmaj_iter = multivar_vmaj_geno_buf
            for sample_batch_idx in range(sample_batch_ct):
                if sample_batch_idx == sample_batch_ct - 1:
                    sample_batch_size = 1 + <uint32_t>((subset_size - 1) % kPglQuaterTransposeBatch)
                smaj_iter = multivar_smaj_geno_batch_buf
                transpose_quaterblock(vmaj_iter, sample_ctaw2, kPglQuaterTransposeWords, variant_batch_size, sample_batch_size, smaj_iter, transpose_batch_buf)
                for uii in range(sample_batch_size):
                    data_ptr = &(geno_int64_out[uii + sample_batch_idx * kPglQuaterTransposeBatch, variant_batch_idx * kPglQuaterTransposeBatch])
                    genoarr_to_int64s_minus9(smaj_iter, variant_batch_size, data_ptr)
                    smaj_iter = &(smaj_iter[kPglQuaterTransposeWords])
                vmaj_iter = &(vmaj_iter[kPglQuaterTransposeWords])
            variant_list_idx += kPglQuaterTransposeBatch
        return

    cpdef read_list(self, np.ndarray[np.uint32_t] variant_idxs, np.ndarray geno_int_out, uint32_t allele_idx = 1, bint sample_maj = 0):
        if geno_int_out.dtype == np.int8:
            self.read_list_internal8(variant_idxs, geno_int_out, allele_idx, sample_maj)
        elif geno_int_out.dtype == np.int32:
            self.read_list_internal32(variant_idxs, geno_int_out, allele_idx, sample_maj)
        elif geno_int_out.dtype == np.int64:
            self.read_list_internal64(variant_idxs, geno_int_out, allele_idx, sample_maj)
        else:
            raise RuntimeError("Invalid read_list() geno_int_out array element type (int8, int32, or int64 expected).")
        return


    cpdef read_alleles_range(self, uint32_t variant_idx_start, uint32_t variant_idx_end, np.ndarray[np.int32_t,mode="c",ndim=2] allele_int32_out, bint hap_maj = 0):
        # if hap_maj == False, allele_int32_out must have at least
        #   variant_idx_ct rows, 2 * sample_ct columns
        # if hap_maj == True, allele_int32_out must have at least 2 * sample_ct
        #   rows, variant_idx_ct columns
        if variant_idx_end > self._info_ptr[0].raw_variant_ct:
            raise RuntimeError("read_alleles_range() variant_idx_end too large (" + str(variant_idx_end) + "; only " + str(self._info_ptr[0].raw_variant_ct) + " in file)")
        cdef const uintptr_t* subset_include_vec = self._subset_include_vec
        cdef const uint32_t* subset_cumulative_popcounts = self._subset_cumulative_popcounts
        cdef pgen_reader_t* pgrp = self._state_ptr
        cdef uintptr_t* genovec = self._genovec
        cdef uintptr_t* phasepresent = self._phasepresent
        cdef uintptr_t* phaseinfo = self._phaseinfo
        cdef uint32_t variant_idx_ct = variant_idx_end - variant_idx_start
        cdef uint32_t subset_size = self._subset_size
        cdef int32_t* main_data_ptr
        cdef uint32_t variant_idx
        cdef uint32_t phasepresent_ct
        cdef pglerr_t reterr
        if hap_maj == 0:
            if allele_int32_out.shape[0] < variant_idx_ct:
                raise RuntimeError("Variant-major read_alleles_range() allele_int32_out buffer has too few rows (" + str(allele_int32_out.shape[0]) + "; (variant_idx_end - variant_idx_start) is " + str(variant_idx_ct) + ")")
            if allele_int32_out.shape[1] < 2 * subset_size:
                raise RuntimeError("Variant-major read_alleles_range() allele_int32_out buffer has too few columns (" + str(allele_int32_out.shape[1]) + "; current sample subset has size " + str(subset_size) + ", and column count should be twice that)")
            for variant_idx in range(variant_idx_start, variant_idx_end):
                # upgrade to multiallelic version of this function later
                reterr = pgr_read_refalt1_genovec_hphase_subset_unsafe(subset_include_vec, subset_cumulative_popcounts, subset_size, variant_idx, pgrp, genovec, phasepresent, phaseinfo, &phasepresent_ct)
                if reterr != kPglRetSuccess:
                    raise RuntimeError("read_alleles_range() error " + str(reterr))
                main_data_ptr = <int32_t*>(&(allele_int32_out[(variant_idx - variant_idx_start), 0]))
                genoarr_phased_to_allele_codes(genovec, phasepresent, phaseinfo, subset_size, phasepresent_ct, NULL, main_data_ptr)
            return
        if variant_idx_start >= variant_idx_end:
            raise RuntimeError("read_alleles_range() variant_idx_start >= variant_idx_end (" + str(variant_idx_start) + ", " + str(variant_idx_end) + ")")
        if allele_int32_out.shape[0] < 2 * subset_size:
            raise RuntimeError("Haplotype-major read_alleles_range() allele_int32_out buffer has too few rows (" + str(allele_int32_out.shape[0]) + "; current sample subset has size " + str(subset_size) + ", and row count should be twice that)")
        if allele_int32_out.shape[1] < variant_idx_ct:
            raise RuntimeError("Haplotype-major read_alleles_range() allele_int32_out buffer has too few columns (" + str(allele_int32_out.shape[1]) + "; (variant_idx_end - variant_idx_start) is " + str(variant_idx_ct) + ")")
        cdef uint32_t variant_batch_ct = DIV_UP(variant_idx_ct, kPglQuaterTransposeBatch)
        cdef uint32_t variant_batch_size = kPglQuaterTransposeBatch
        cdef uint32_t variant_batch_sizel = DIV_UP(variant_batch_size, kBitsPerWord)
        cdef uint32_t variant_idx_offset = variant_idx_start
        cdef uint32_t sample_ctaw2 = kWordsPerVec * DIV_UP(subset_size, kBitsPerWordD2)
        cdef uint32_t sample_ctaw = kWordsPerVec * DIV_UP(subset_size, kBitsPerWord)
        cdef uint32_t sample_batch_ct = DIV_UP(subset_size, kPglQuaterTransposeBatch)
        cdef vul_t* transpose_batch_buf = self._transpose_batch_buf
        cdef uintptr_t* multivar_vmaj_geno_buf = self._multivar_vmaj_geno_buf
        cdef uintptr_t* multivar_vmaj_phaseinfo_buf = self._multivar_vmaj_phaseinfo_buf
        cdef uintptr_t* multivar_smaj_geno_batch_buf = self._multivar_smaj_geno_batch_buf
        cdef uintptr_t* multivar_smaj_phaseinfo_batch_buf = self._multivar_smaj_phaseinfo_batch_buf
        cdef uintptr_t* vmaj_geno_iter
        cdef uintptr_t* vmaj_phaseinfo_iter
        cdef uintptr_t* smaj_geno_iter
        cdef uintptr_t* smaj_phaseinfo_iter
        cdef int32_t* main_data1_ptr
        cdef uint32_t variant_batch_idx
        cdef uint32_t sample_batch_size
        cdef uint32_t sample_batch_idx
        cdef uint32_t uii
        for variant_batch_idx in range(variant_batch_ct):
            if variant_batch_idx == (variant_batch_ct - 1):
                variant_batch_size = 1 + <uint32_t>((variant_idx_ct - 1) % kPglQuaterTransposeBatch)
                variant_batch_sizel = DIV_UP(variant_batch_size, kBitsPerWord)
            vmaj_geno_iter = multivar_vmaj_geno_buf
            vmaj_phaseinfo_iter = multivar_vmaj_phaseinfo_buf
            for uii in range(variant_batch_size):
                reterr = pgr_read_refalt1_genovec_hphase_subset_unsafe(subset_include_vec, subset_cumulative_popcounts, subset_size, uii + variant_idx_offset, pgrp, vmaj_geno_iter, phasepresent, vmaj_phaseinfo_iter, &phasepresent_ct)
                if reterr != kPglRetSuccess:
                    raise RuntimeError("read_alleles_range() error " + str(reterr))
                if phasepresent_ct == 0:
                    fill_ulong_zero(sample_ctaw, vmaj_phaseinfo_iter)
                # else:
                    # bitvec_and(phasepresent, sample_ctaw, vmaj_phaseinfo_iter)
                vmaj_geno_iter = &(vmaj_geno_iter[sample_ctaw2])
                vmaj_phaseinfo_iter = &(vmaj_phaseinfo_iter[sample_ctaw])
            sample_batch_size = kPglQuaterTransposeBatch
            vmaj_geno_iter = multivar_vmaj_geno_buf
            vmaj_phaseinfo_iter = multivar_vmaj_phaseinfo_buf
            for sample_batch_idx in range(sample_batch_ct):
                if sample_batch_idx == sample_batch_ct - 1:
                    sample_batch_size = 1 + <uint32_t>((subset_size - 1) % kPglQuaterTransposeBatch)
                smaj_geno_iter = multivar_smaj_geno_batch_buf
                smaj_phaseinfo_iter = multivar_smaj_phaseinfo_batch_buf
                transpose_quaterblock(vmaj_geno_iter, sample_ctaw2, kPglQuaterTransposeWords, variant_batch_size, sample_batch_size, smaj_geno_iter, transpose_batch_buf)
                # todo: skip bitblock transpose when all phasepresent_ct values
                #       are zero, etc.
                transpose_bitblock(vmaj_phaseinfo_iter, sample_ctaw, <uint32_t>(kPglQuaterTransposeWords / 2), variant_batch_size, sample_batch_size, smaj_phaseinfo_iter, transpose_batch_buf)
                for uii in range(sample_batch_size):
                    main_data_ptr = <int32_t*>(&(allele_int32_out[2 * (uii + sample_batch_idx * kPglQuaterTransposeWords), variant_batch_idx * kPglQuaterTransposeBatch]))
                    main_data1_ptr = <int32_t*>(&(allele_int32_out[2 * (uii + sample_batch_idx * kPglQuaterTransposeWords) + 1, variant_batch_idx * kPglQuaterTransposeBatch]))
                    genoarr_phased_to_hap_codes(smaj_geno_iter, smaj_phaseinfo_iter, variant_batch_size, main_data_ptr, main_data1_ptr)
                    smaj_geno_iter = &(smaj_geno_iter[kPglQuaterTransposeWords])
                    smaj_phaseinfo_iter = &(smaj_phaseinfo_iter[kPglQuaterTransposeWords / 2])
                vmaj_geno_iter = &(vmaj_geno_iter[kPglQuaterTransposeWords])
                vmaj_phaseinfo_iter = &(vmaj_phaseinfo_iter[kPglQuaterTransposeWords / 2])
            variant_idx_offset += kPglQuaterTransposeBatch
        return


    cpdef read_alleles_list(self, np.ndarray[np.uint32_t] variant_idxs, np.ndarray[np.int32_t,mode="c",ndim=2] allele_int32_out, bint hap_maj = 0):
        # if hap_maj == False, allele_int32_out must have at least
        #   variant_idx_ct rows, 2 * sample_ct columns
        # if hap_maj == True, allele_int32_out must have at least 2 * sample_ct
        #   rows, variant_idx_ct columns
        cdef uint32_t raw_variant_ct = self._info_ptr[0].raw_variant_ct
        cdef const uintptr_t* subset_include_vec = self._subset_include_vec
        cdef const uint32_t* subset_cumulative_popcounts = self._subset_cumulative_popcounts
        cdef pgen_reader_t* pgrp = self._state_ptr
        cdef uintptr_t* genovec = self._genovec
        cdef uintptr_t* phasepresent = self._phasepresent
        cdef uintptr_t* phaseinfo = self._phaseinfo
        cdef uint32_t variant_idx_ct = <uint32_t>variant_idxs.shape[0]
        cdef uint32_t subset_size = self._subset_size
        cdef int32_t* main_data_ptr
        cdef uint32_t variant_list_idx
        cdef uint32_t variant_idx
        cdef uint32_t phasepresent_ct
        cdef pglerr_t reterr
        if hap_maj == 0:
            if allele_int32_out.shape[0] < variant_idx_ct:
                raise RuntimeError("Variant-major read_alleles_list() allele_int32_out buffer has too few rows (" + str(allele_int32_out.shape[0]) + "; variant_idxs length is " + str(variant_idx_ct) + ")")
            if allele_int32_out.shape[1] < 2 * subset_size:
                raise RuntimeError("Variant-major read_alleles_list() allele_int32_out buffer has too few columns (" + str(allele_int32_out.shape[1]) + "; current sample subset has size " + str(subset_size) + ", and column count should be twice that)")
            for variant_list_idx in range(variant_idx_ct):
                variant_idx = variant_idxs[variant_list_idx]
                if variant_idx >= raw_variant_ct:
                    raise RuntimeError("read_alleles_list() variant index too large (" + str(variant_idx) + "; only " + str(raw_variant_ct) + " in file)")
                # upgrade to multiallelic version of this function later
                reterr = pgr_read_refalt1_genovec_hphase_subset_unsafe(subset_include_vec, subset_cumulative_popcounts, subset_size, variant_idx, pgrp, genovec, phasepresent, phaseinfo, &phasepresent_ct)
                if reterr != kPglRetSuccess:
                    raise RuntimeError("read_alleles_list() error " + str(reterr))
                main_data_ptr = <int32_t*>(&(allele_int32_out[variant_list_idx, 0]))
                genoarr_phased_to_allele_codes(genovec, phasepresent, phaseinfo, subset_size, phasepresent_ct, NULL, main_data_ptr)
            return
        if allele_int32_out.shape[0] < 2 * subset_size:
            raise RuntimeError("Haplotype-major read_alleles_list() allele_int32_out buffer has too few rows (" + str(allele_int32_out.shape[0]) + "; current sample subset has size " + str(subset_size) + ", and row count should be twice that)")
        if allele_int32_out.shape[1] < variant_idx_ct:
            raise RuntimeError("Haplotype-major read_alleles_list() allele_int32_out buffer has too few columns (" + str(allele_int32_out.shape[1]) + "; variant_idxs length is " + str(variant_idx_ct) + ")")
        cdef uint32_t variant_batch_ct = DIV_UP(variant_idx_ct, kPglQuaterTransposeBatch)
        cdef uint32_t variant_batch_size = kPglQuaterTransposeBatch
        cdef uint32_t variant_batch_sizel = DIV_UP(variant_batch_size, kBitsPerWord)
        cdef uint32_t sample_ctaw2 = kWordsPerVec * DIV_UP(subset_size, kBitsPerWordD2)
        cdef uint32_t sample_ctaw = kWordsPerVec * DIV_UP(subset_size, kBitsPerWord)
        cdef uint32_t sample_batch_ct = DIV_UP(subset_size, kPglQuaterTransposeBatch)
        cdef vul_t* transpose_batch_buf = self._transpose_batch_buf
        cdef uintptr_t* multivar_vmaj_geno_buf = self._multivar_vmaj_geno_buf
        cdef uintptr_t* multivar_vmaj_phaseinfo_buf = self._multivar_vmaj_phaseinfo_buf
        cdef uintptr_t* multivar_smaj_geno_batch_buf = self._multivar_smaj_geno_batch_buf
        cdef uintptr_t* multivar_smaj_phaseinfo_batch_buf = self._multivar_smaj_phaseinfo_batch_buf
        cdef uintptr_t* vmaj_geno_iter
        cdef uintptr_t* vmaj_phaseinfo_iter
        cdef uintptr_t* smaj_geno_iter
        cdef uintptr_t* smaj_phaseinfo_iter
        cdef int32_t* main_data1_ptr
        cdef uint32_t variant_batch_idx
        cdef uint32_t sample_batch_size
        cdef uint32_t sample_batch_idx
        cdef uint32_t uii
        for variant_batch_idx in range(variant_batch_ct):
            if variant_batch_idx == (variant_batch_ct - 1):
                variant_batch_size = 1 + <uint32_t>((variant_idx_ct - 1) % kPglQuaterTransposeBatch)
                variant_batch_sizel = DIV_UP(variant_batch_size, kBitsPerWord)
            vmaj_geno_iter = multivar_vmaj_geno_buf
            vmaj_phaseinfo_iter = multivar_vmaj_phaseinfo_buf
            for variant_list_idx in range(variant_batch_idx * kPglQuaterTransposeBatch, variant_batch_idx * kPglQuaterTransposeBatch + variant_batch_size):
                variant_idx = variant_idxs[variant_list_idx]
                if variant_idx >= raw_variant_ct:
                    raise RuntimeError("read_alleles_list() variant index too large (" + str(variant_idx) + "; only " + str(raw_variant_ct) + " in file)")
                reterr = pgr_read_refalt1_genovec_hphase_subset_unsafe(subset_include_vec, subset_cumulative_popcounts, subset_size, variant_idx, pgrp, vmaj_geno_iter, phasepresent, vmaj_phaseinfo_iter, &phasepresent_ct)
                if reterr != kPglRetSuccess:
                    raise RuntimeError("read_alleles_list() error " + str(reterr))
                if phasepresent_ct == 0:
                    fill_ulong_zero(sample_ctaw, vmaj_phaseinfo_iter)
                # else:
                    # bitvec_and(phasepresent, sample_ctaw, vmaj_phaseinfo_iter)
                vmaj_geno_iter = &(vmaj_geno_iter[sample_ctaw2])
                vmaj_phaseinfo_iter = &(vmaj_phaseinfo_iter[sample_ctaw])
            sample_batch_size = kPglQuaterTransposeBatch
            vmaj_geno_iter = multivar_vmaj_geno_buf
            vmaj_phaseinfo_iter = multivar_vmaj_phaseinfo_buf
            for sample_batch_idx in range(sample_batch_ct):
                if sample_batch_idx == sample_batch_ct - 1:
                    sample_batch_size = 1 + <uint32_t>((subset_size - 1) % kPglQuaterTransposeBatch)
                smaj_geno_iter = multivar_smaj_geno_batch_buf
                smaj_phaseinfo_iter = multivar_smaj_phaseinfo_batch_buf
                transpose_quaterblock(vmaj_geno_iter, sample_ctaw2, kPglQuaterTransposeWords, variant_batch_size, sample_batch_size, smaj_geno_iter, transpose_batch_buf)
                # todo: skip bitblock transpose when all phasepresent_ct values
                #       are zero, etc.
                transpose_bitblock(vmaj_phaseinfo_iter, sample_ctaw, <uint32_t>(kPglQuaterTransposeWords / 2), variant_batch_size, sample_batch_size, smaj_phaseinfo_iter, transpose_batch_buf)
                for uii in range(sample_batch_size):
                    main_data_ptr = <int32_t*>(&(allele_int32_out[2 * (uii + sample_batch_idx * kPglQuaterTransposeWords), variant_batch_idx * kPglQuaterTransposeBatch]))
                    main_data1_ptr = <int32_t*>(&(allele_int32_out[2 * (uii + sample_batch_idx * kPglQuaterTransposeWords) + 1, variant_batch_idx * kPglQuaterTransposeBatch]))
                    genoarr_phased_to_hap_codes(smaj_geno_iter, smaj_phaseinfo_iter, variant_batch_size, main_data_ptr, main_data1_ptr)
                    smaj_geno_iter = &(smaj_geno_iter[kPglQuaterTransposeWords])
                    smaj_phaseinfo_iter = &(smaj_phaseinfo_iter[kPglQuaterTransposeWords / 2])
                vmaj_geno_iter = &(vmaj_geno_iter[kPglQuaterTransposeWords])
                vmaj_phaseinfo_iter = &(vmaj_phaseinfo_iter[kPglQuaterTransposeWords / 2])
        return


    cpdef read_alleles_and_phasepresent_range(self, uint32_t variant_idx_start, uint32_t variant_idx_end, np.ndarray[np.int32_t,mode="c",ndim=2] allele_int32_out, np.ndarray[np.uint8_t,mode="c",cast=True,ndim=2] phasepresent_out, bint hap_maj = 0):
        pass


    cpdef read_alleles_and_phasepresent_list(self, np.ndarray[np.uint32_t] variant_idxs, np.ndarray[np.int32_t,mode="c",ndim=2] allele_int32_out, np.ndarray[np.uint8_t,cast=True,mode="c",ndim=2] phasepresent_out, bint hap_maj = 0):
        pass

    
    cpdef count(self, uint32_t variant_idx, np.ndarray[np.uint32_t,mode="c"] genocount_uint32_out, object allele_idx = 1):
        # todo: multiallelic variants
        if allele_idx is None:
            allele_idx = 1
        cdef uint32_t* data_ptr = <uint32_t*>(&(genocount_uint32_out[0]))
        cdef pglerr_t reterr = pgr_get_refalt1_genotype_counts(self._subset_include_vec, self._subset_include_interleaved_vec, self._subset_cumulative_popcounts, self._subset_size, variant_idx, self._state_ptr, data_ptr)
        if reterr != kPglRetSuccess:
            raise RuntimeError("count() error " + str(reterr))
        if allele_idx != 0:
            return
        cdef uint32_t tmp = data_ptr[0]
        data_ptr[0] = data_ptr[2]
        data_ptr[2] = tmp
        return

    
    cpdef change_sample_subset(self, object sample_subset = None):
        if sample_subset is not None:
            self.set_sample_subset_internal(sample_subset)
        else:
            self._subset_size = self._info_ptr[0].raw_sample_ct
        return

    
    cpdef close(self):
        # don't bother propagating file close errors for now
        if self._info_ptr:
            pgfi_cleanup(self._info_ptr)
            if self._info_ptr[0].vrtypes:
                aligned_free(self._info_ptr[0].vrtypes)
                if self._state_ptr:
                    pgr_cleanup(self._state_ptr)
                    if self._state_ptr[0].fread_buf:
                        aligned_free(self._state_ptr[0].fread_buf)
                    PyMem_Free(self._state_ptr)
                    self._state_ptr = NULL
            PyMem_Free(self._info_ptr)
            self._info_ptr = NULL
        return

    
    cpdef __exit__(self, exc_type, exc_val, exc_tb):
        self.close()
        return

    
    def __dealloc__(self):
        if self._info_ptr:
            pgfi_cleanup(self._info_ptr)
            if self._info_ptr[0].vrtypes:
                aligned_free(self._info_ptr[0].vrtypes)
                if self._state_ptr:
                    pgr_cleanup(self._state_ptr)
                    if self._state_ptr[0].fread_buf:
                        aligned_free(self._state_ptr[0].fread_buf)
                    PyMem_Free(self._state_ptr)
            PyMem_Free(self._info_ptr)
        return



cdef bytes_to_bits_internal(np.ndarray[np.uint8_t,mode="c",cast=True] boolbytes, uint32_t sample_ct, uintptr_t* bitarr):
    bytes_to_bits_unsafe(boolbytes, sample_ct, bitarr)

cdef class PgenWriter:
    cdef st_pgen_writer_t* _state_ptr
    cdef uintptr_t* _nonref_flags
    # preallocate buffers we'll use repeatedly
    cdef uintptr_t* _genovec
    cdef uintptr_t* _phasepresent
    cdef uintptr_t* _phaseinfo
    cdef uintptr_t* _dosage_present
    cdef uint16_t* _dosage_vals
    
    
    def __cinit__(self, bytes filename, uint32_t sample_ct,
                  uint32_t variant_ct, object nonref_flags,
                  object allele_idx_offsets = None,
                  bint hardcall_phase_present = False,
                  bint dosage_present = False,
                  bint dosage_phase_present = False):
        if dosage_phase_present and not dosage_present:
            raise RuntimeError("Invalid arguments for PgenWriter constructor (dosage_phase_present true but dosage_present false).")
        if allele_idx_offsets is not None:
            for uii in range(variant_ct + 1):
                if allele_idx_offsets[uii] != uii * 2:
                    raise RuntimeError("Multiallelic variants aren't supported by PgenWriter yet.")
                
        self._state_ptr = <st_pgen_writer_t*>PyMem_Malloc(sizeof(st_pgen_writer_t))
        if not self._state_ptr:
            raise MemoryError()
        self._nonref_flags = NULL
        cdef uint32_t nonref_flags_storage = 0
        cdef uint32_t bitvec_cacheline_ct = DIV_UP(sample_ct, kBitsPerCacheline)
        if nonref_flags is not None:
            if type(nonref_flags) == type(True):
                if nonref_flags:
                    nonref_flags_storage = 2
                else:
                    nonref_flags_storage = 1
            else:
                nonref_flags_storage = 3
                if cachealigned_malloc(bitvec_cacheline_ct * kCacheline, &(self._nonref_flags)):
                    raise MemoryError()
                bytes_to_bits_internal(nonref_flags, sample_ct, self._nonref_flags)
        cdef const char* fname = <const char*>filename
        cdef pgen_global_flags_t phase_dosage_gflags = kfPgenGlobal0
        if hardcall_phase_present:
            phase_dosage_gflags |= kfPgenGlobalHardcallPhasePresent
        if dosage_present:
            phase_dosage_gflags |= kfPgenGlobalDosagePresent
        assert not dosage_phase_present
        cdef uintptr_t alloc_cacheline_ct
        cdef uint32_t max_vrec_len
        cdef pglerr_t reterr = spgw_init_phase1(fname, NULL, self._nonref_flags, variant_ct, sample_ct, phase_dosage_gflags, nonref_flags_storage, self._state_ptr, &alloc_cacheline_ct, &max_vrec_len)
        if reterr != kPglRetSuccess:
            raise RuntimeError("spgw_init_phase1() error " + str(reterr))
        cdef uint32_t genovec_cacheline_ct = DIV_UP(sample_ct, kQuatersPerCacheline)
        cdef uint32_t dosage_vals_cacheline_ct = DIV_UP(sample_ct, (2 * kInt32PerCacheline))
        cdef unsigned char* spgw_alloc
        if cachealigned_malloc((alloc_cacheline_ct + genovec_cacheline_ct + 3 * bitvec_cacheline_ct + dosage_vals_cacheline_ct) * kCacheline, &spgw_alloc):
            raise MemoryError()
        spgw_init_phase2(max_vrec_len, self._state_ptr, spgw_alloc)  
        self._genovec = <uintptr_t*>(&(spgw_alloc[alloc_cacheline_ct * kCacheline]))
        self._phasepresent = <uintptr_t*>(&(spgw_alloc[(alloc_cacheline_ct + genovec_cacheline_ct) * kCacheline]))
        self._phaseinfo = <uintptr_t*>(&(spgw_alloc[(alloc_cacheline_ct + genovec_cacheline_ct + bitvec_cacheline_ct) * kCacheline]))
        self._dosage_present = <uintptr_t*>(&(spgw_alloc[(alloc_cacheline_ct + genovec_cacheline_ct + 2 * bitvec_cacheline_ct) * kCacheline]))
        self._dosage_vals = <uint16_t*>(&(spgw_alloc[(alloc_cacheline_ct + genovec_cacheline_ct + 3 * bitvec_cacheline_ct) * kCacheline]))
        return

    
    cpdef __enter__(self):
        return self


    cpdef append_biallelic(self, np.ndarray[np.int8_t,mode="c"] geno_int8):
        cdef int8_t* genobytes = &(geno_int8[0])
        bytes_to_genoarr_unsafe(genobytes, self._state_ptr[0].pwc.sample_ct, self._genovec)
        cdef pglerr_t reterr = spgw_append_biallelic_genovec(self._genovec, self._state_ptr)
        if reterr != kPglRetSuccess:
            raise RuntimeError("append_biallelic() error " + str(reterr))
        return
    

    cpdef append_alleles(self, np.ndarray[np.int32_t,mode="c"] allele_int32, bint all_phased = False):
        cdef int32_t* allele_codes = <int32_t*>(&(allele_int32[0]))
        cdef uintptr_t* genovec = self._genovec
        cdef pglerr_t reterr
        if not all_phased:
            allele_codes_to_genoarr_unsafe(allele_codes, NULL, self._state_ptr[0].pwc.sample_ct, genovec, NULL, NULL)
            reterr = spgw_append_biallelic_genovec(genovec, self._state_ptr)
        else:
            allele_codes_to_genoarr_unsafe(allele_codes, NULL, self._state_ptr[0].pwc.sample_ct, genovec, self._phasepresent, self._phaseinfo)
            reterr = spgw_append_biallelic_genovec_hphase(genovec, self._phasepresent, self._phaseinfo, self._state_ptr)
        if reterr != kPglRetSuccess:
            raise RuntimeError("append_alleles() error " + str(reterr))
        return

    
    cpdef append_partially_phased(self, np.ndarray[np.int32_t,mode="c"] allele_int32, np.ndarray[np.uint8_t,cast=True] phasepresent):
        cdef int32_t* allele_codes = <int32_t*>(&(allele_int32[0]))
        cdef unsigned char* phasepresent_bytes = <unsigned char*>(&(phasepresent[0]))
        cdef uintptr_t* genovec = self._genovec
        cdef uintptr_t* phasepresent_buf = self._phasepresent
        cdef uintptr_t* phaseinfo = self._phaseinfo
        allele_codes_to_genoarr_unsafe(allele_codes, phasepresent_bytes, self._state_ptr[0].pwc.sample_ct, genovec, phasepresent_buf, phaseinfo)
        cdef pglerr_t reterr = spgw_append_biallelic_genovec_hphase(genovec, phasepresent_buf, phaseinfo, self._state_ptr)
        if reterr != kPglRetSuccess:
            raise RuntimeError("append_partially_phased() error " + str(reterr))
        return


    cdef append_dosages_internal32(self, np.ndarray[np.float32_t,mode="c"] floatarr):
        cdef uintptr_t* genovec = self._genovec
        cdef uintptr_t* dosage_present = self._dosage_present
        cdef uint16_t* dosage_vals = self._dosage_vals
        cdef uint32_t dosage_ct
        floats_to_dosage16(<float*>(&(floatarr[0])), self._state_ptr[0].pwc.sample_ct, 6554, genovec, dosage_present, dosage_vals, &dosage_ct)
        cdef pglerr_t reterr = spgw_append_biallelic_genovec_dosage16(genovec, dosage_present, dosage_vals, dosage_ct, self._state_ptr)
        if reterr != kPglRetSuccess:
            raise RuntimeError("append_dosages() error " + str(reterr))
        return

    cdef append_dosages_internal64(self, np.ndarray[np.float64_t,mode="c"] doublearr):
        cdef uintptr_t* genovec = self._genovec
        cdef uintptr_t* dosage_present = self._dosage_present
        cdef uint16_t* dosage_vals = self._dosage_vals
        cdef uint32_t dosage_ct
        doubles_to_dosage16(<double*>(&(doublearr[0])), self._state_ptr[0].pwc.sample_ct, 6554, genovec, dosage_present, dosage_vals, &dosage_ct)
        cdef pglerr_t reterr = spgw_append_biallelic_genovec_dosage16(genovec, dosage_present, dosage_vals, dosage_ct, self._state_ptr)
        if reterr != kPglRetSuccess:
            raise RuntimeError("append_dosages() error " + str(reterr))
        return

    cpdef append_dosages(self, np.ndarray floatarr):
        if floatarr.dtype == np.float32:
            self.append_dosages_internal32(floatarr)
        elif floatarr.dtype == np.float64:
            self.append_dosages_internal64(floatarr)
        else:
            raise RuntimeError("Invalid append_dosages() dosage array element type (float32 or float64 expected).")
        return


    cpdef append_biallelic_batch(self, np.ndarray[np.int8_t,mode="c",ndim=2] geno_int8_batch):
        cdef uint32_t batch_size = <uint32_t>geno_int8_batch.shape[0]
        cdef int8_t* genobytes
        cdef uint32_t uii
        cdef pglerr_t reterr
        for uii in range(batch_size):
            genobytes = &(geno_int8_batch[uii, 0])
            bytes_to_genoarr_unsafe(genobytes, self._state_ptr[0].pwc.sample_ct, self._genovec)
            reterr = spgw_append_biallelic_genovec(self._genovec, self._state_ptr)
            if reterr != kPglRetSuccess:
                raise RuntimeError("append_biallelic_batch() error " + str(reterr))
        return
    

    cpdef append_alleles_batch(self, np.ndarray[np.int32_t,mode="c",ndim=2] allele_int32_batch, bint all_phased = False):
        cdef uint32_t batch_size = <uint32_t>allele_int32_batch.shape[0]
        cdef uintptr_t* genovec = self._genovec
        cdef int32_t* allele_codes
        cdef uint32_t uii
        cdef pglerr_t reterr
        if not all_phased:
            for uii in range(batch_size):
                allele_codes = <int32_t*>(&(allele_int32_batch[uii, 0]))
                allele_codes_to_genoarr_unsafe(allele_codes, NULL, self._state_ptr[0].pwc.sample_ct, genovec, NULL, NULL)
                reterr = spgw_append_biallelic_genovec(genovec, self._state_ptr)
                if reterr != kPglRetSuccess:
                    raise RuntimeError("append_alleles_batch() error " + str(reterr))
        else:
            for uii in range(batch_size):
                allele_codes = <int32_t*>(&(allele_int32_batch[uii, 0]))
                allele_codes_to_genoarr_unsafe(allele_codes, NULL, self._state_ptr[0].pwc.sample_ct, genovec, self._phasepresent, self._phaseinfo)
                reterr = spgw_append_biallelic_genovec_hphase(genovec, self._phasepresent, self._phaseinfo, self._state_ptr)
                if reterr != kPglRetSuccess:
                    raise RuntimeError("append_alleles_batch() error " + str(reterr))
        return

    
    cpdef append_partially_phased_batch(self, np.ndarray[np.int32_t,mode="c",ndim=2] allele_int32_batch, np.ndarray[np.uint8_t,mode="c",cast=True,ndim=2] phasepresent_batch):
        cdef uint32_t batch_size = <uint32_t>allele_int32_batch.shape[0]
        cdef uintptr_t* genovec = self._genovec
        cdef uintptr_t* phasepresent_buf = self._phasepresent
        cdef uintptr_t* phaseinfo = self._phaseinfo
        cdef int32_t* allele_codes
        cdef unsigned char* phasepresent_bytes
        cdef uint32_t uii
        cdef pglerr_t reterr
        for uii in range(batch_size):
            allele_codes = <int32_t*>(&(allele_int32_batch[uii, 0]))
            phasepresent_bytes = <unsigned char*>(&(phasepresent_batch[uii, 0]))
            allele_codes_to_genoarr_unsafe(allele_codes, phasepresent_bytes, self._state_ptr[0].pwc.sample_ct, genovec, phasepresent_buf, phaseinfo)
            reterr = spgw_append_biallelic_genovec_hphase(genovec, phasepresent_buf, phaseinfo, self._state_ptr)
            if reterr != kPglRetSuccess:
                raise RuntimeError("append_partially_phased_batch() error " + str(reterr))
        return


    cdef append_dosages_batch_internal32(self, np.ndarray[np.float32_t,mode="c",ndim=2] floatarr_batch):
        cdef uint32_t batch_size = <uint32_t>floatarr_batch.shape[0]
        cdef uintptr_t* genovec = self._genovec
        cdef uintptr_t* dosage_present = self._dosage_present
        cdef uint16_t* dosage_vals = self._dosage_vals
        cdef uint32_t dosage_ct
        cdef uint32_t uii
        cdef pglerr_t reterr
        for uii in range(batch_size):
            floats_to_dosage16(<float*>(&(floatarr_batch[uii, 0])), self._state_ptr[0].pwc.sample_ct, 6554, genovec, dosage_present, dosage_vals, &dosage_ct)
            reterr = spgw_append_biallelic_genovec_dosage16(genovec, dosage_present, dosage_vals, dosage_ct, self._state_ptr)
            if reterr != kPglRetSuccess:
                raise RuntimeError("append_dosages_batch() error " + str(reterr))
        return

    cdef append_dosages_batch_internal64(self, np.ndarray[np.float64_t,mode="c",ndim=2] doublearr_batch):
        cdef uint32_t batch_size = <uint32_t>doublearr_batch.shape[0]
        cdef uintptr_t* genovec = self._genovec
        cdef uintptr_t* dosage_present = self._dosage_present
        cdef uint16_t* dosage_vals = self._dosage_vals
        cdef uint32_t dosage_ct
        cdef uint32_t uii
        cdef pglerr_t reterr
        for uii in range(batch_size):
            doubles_to_dosage16(<double*>(&(doublearr_batch[uii, 0])), self._state_ptr[0].pwc.sample_ct, 6554, genovec, dosage_present, dosage_vals, &dosage_ct)
            reterr = spgw_append_biallelic_genovec_dosage16(genovec, dosage_present, dosage_vals, dosage_ct, self._state_ptr)
            if reterr != kPglRetSuccess:
                raise RuntimeError("append_dosages_batch() error " + str(reterr))
        return

    cpdef append_dosages_batch(self, np.ndarray floatarr_batch):
        if floatarr_batch.dtype == np.float32:
            self.append_dosages_batch_internal32(floatarr_batch)
        elif floatarr_batch.dtype == np.float64:
            self.append_dosages_batch_internal64(floatarr_batch)
        else:
            raise RuntimeError("Invalid append_dosages_batch() dosage array element type (float32 or float64 expected).")
        return
    
    
    cpdef close(self):
        if self._state_ptr:
            if self._state_ptr[0].pwc.vidx != self._state_ptr[0].pwc.variant_ct:
                raise RuntimeError("PgenWriter.close() called when number of written variants (" + str(self._state_ptr[0].pwc.vidx) + ") unequal to initially declared value (" + str(self._state_ptr[0].pwc.variant_ct) + ").")
            spgw_finish(self._state_ptr)
            if self._nonref_flags:
                aligned_free(self._nonref_flags)
            PyMem_Free(self._state_ptr)
            self._state_ptr = NULL
        return


    cpdef __exit__(self, exc_type, exc_val, exc_tb):
        self.close()
        return


    def __dealloc__(self):
        if self._state_ptr:
            if self._state_ptr[0].pwc.vidx == self._state_ptr[0].pwc.variant_ct:
                spgw_finish(self._state_ptr)
            else:
                spgw_cleanup(self._state_ptr)
            if self._nonref_flags:
                aligned_free(self._nonref_flags)
            PyMem_Free(self._state_ptr)
        return



import datetime as dt

class Stepper(object):
    def __init__(self, prefix=''):
        """

        :type prefix: basestring
        """
        self.start = dt.datetime.now()
        self.ncalls = 0
        self.prefix = prefix
        self.write(msg="Start @ "+dt.datetime.now().strftime("%Y-%m-%d %H:%M:%S"))

    @staticmethod
    def formatTD(td):
        hours = td.seconds // 3600
        minutes = (td.seconds % 3600) // 60
        seconds = td.seconds % 60
        return '%02d:%02d:%02d' % (hours, minutes, seconds)

    def new(self, msg='Step'):
        self.ncalls += 1
        print(self.prefix, self.ncalls, ". [", self.formatTD(dt.datetime.now()-self.start), "] ", msg)
        return dt.datetime.now()-self.start

    def write(self, msg):
        print ((10+len(self.prefix))*' '+msg)

    def end(self):
        return self.new(msg="The End @ "+dt.datetime.now().strftime("%Y-%m-%d %H:%M:%S"))




##from __future__ import division
from operator import itemgetter
from bisect import *
import sys
import numpy as np
##import numpy.ma as ma
from hapflk import missing, complete_cases, is_na

zero=np.int16(0)
un=np.int16(1)
deux=np.int16(2)

class Dataset():
    def __init__(self,fileName,nsnp,nindiv):
        self.name=fileName
        ## Individuals
        self.nindiv=nindiv
        self.iindiv=0
        self.indiv={}
        ## Individual index in Data Matrix
        self.indivIdx={}
        ## SNPs
        self.nsnp=nsnp
        self.isnp=0
        self.snp={}
        ## SNP indices in Data Matrix
        self.snpIdx={}
        self._initDataMatrix()
        ## populations
        ## populations are accessed by a name (key of the dict)
        ## the value is a vector of booleans taking a True value for individuals
        ## belonging to the pop otherwise
        ## These vectors can be used for fast access to population genotypes e.g.:
        ## vpop=self.populations['mypop']
        ## mypop_genotypes=self.Data[vpop,]
        self.populations={}
    def addSnp(self,Name):
        ''' Add a snp to the dataset '''
        if Name in self.snp:
            return
        self.snp[Name]=SNP(Name)
        self.snpIdx[Name]=self.isnp
        self.isnp+=1
    def rmSnp(self,Name):
        ''' 
        Remove a snp from the dataset 
        The SNP is made unaccessible but the data stay in the Data Matrix
        '''
        if not Name in self.snp:
            return
        else:
            del self.snp[Name]
            del self.snpIdx[Name]
            del self._allSnps
    def addIndividual(self,ID,pop=None,sex=None,fatherID=None,motherID=None,phenotype=None):
        ''' Add an individual to the dataset'''
        if ID in self.indiv:
            print( "Warning: Individual ",ID," has been seen before !")
            return
        self.indiv[ID]=Individual(ID,pop,sex,fatherID,motherID,phenotype)
        self.indivIdx[ID]=self.iindiv
        if pop is not None:
            try:
                vpop=self.populations[pop]
            except KeyError:
                vpop=np.array([0]*self.nindiv,dtype=bool)
                self.populations[pop]=vpop
            vpop[self.iindiv]=True
        self.iindiv+=1
    def _initDataMatrix(self):
        ''' Initialize Data Matrix to an int16 array filled with -1'''
        self.Data=-np.ones(shape=(self.nindiv,self.nsnp),dtype='int16')
    def IndivGenotype(self,indiv,snps=None):
        '''
        Return (0,1,2) genotype of individual INDIV at all SNPs in the dataset
        '''
        try:
            i=self.indivIdx[indiv]
        except KeyError:
            print( 'Individual',indiv,'not in Data')
            raise
        if snps is None:
            return [ (x<0 and missing) or x for x in  self.Data[i,]]
        else:
            return [ (x<0 and missing) or x for x in self.Data[i,[self.snpIdx[s] for s in snps]]]
    def IndivGenotype_char(self,indiv,snps=None):
        igeno=self.IndivGenotype(indiv,snps)
        cgeno=[]
        if snps is None:
            snps=self.AllSnps()
        for i,s in enumerate(snps):
            cgeno.append(self.Alleles(igeno[i],s))
        return cgeno
    def AllSnps(self):
        try:
            return self._allSnps
        except AttributeError:
            self._allSnps=sorted(self.snp,key=lambda x: self.snpIdx[x])
        return self._allSnps
    def SnpGenotype(self,snp):
        '''
        Return genotype of snp SNP for all individuals in the dataset
        '''
        try:
            j=self.snpIdx[snp]
        except KeyError:
            return [missing]*self.nindiv
        return [ (x<0 and missing) or x for x in self.Data[:,j]]
    def Alleles(self,geno,snp):
        '''
        Return the alleles corresponding to genotype GENO at snp SNP
        '''
        if geno is missing:
            return ('?','?')
        snpAll=self.snp[snp].alleles
        if geno==0:
            return (snpAll[0],snpAll[0])
        elif geno==1:
            return (snpAll[0],snpAll[1])
        else:
            assert geno==2
            return (snpAll[1],snpAll[1])
    def Genotype(self,indiv,snp):
        '''
        Return the genotype of individual INDIV  at snp SNP 
        If SNP is not found in the dataset, returns 'missing'
        '''
        try:
            i=self.indivIdx[indiv]
        except KeyError:
            print( 'Individual',indiv,'not in Data')
            raise
        try:
            j=self.snpIdx[snp]
        except KeyError:
            return missing
        return self.Data[i,j]<0 and missing or self.Data[i,j]
    def SetGenotypeSlow(self,indiv,snp,value):
        '''
        Set the genotype of individual INDIV at snp SNP to value VALUE
        Correct Genotype is a value of length 2, with alleles matching the SNP
        '''
        try:
            i=self.indivIdx[indiv]
        except KeyError:
            print( 'Individual',indiv,'not in Data')
            raise
        try:
            j=self.snpIdx[snp]
        except KeyError:
            print( 'SNP',snp,'not in Data')
            raise
        if self.Data[i,j] > -1:
            self.snp[snp].delGenotype(self.Data[i,j])
        geno=self.snp[snp].addObservation(value)
        if geno is not missing:
            self.indiv[indiv].callrate+=1
        self.Data[i,j]=geno
    def SetGenotype(self,indiv,snp,value):
        '''
        Fast Version of SetGenotype with no checking for errors 
        Use with caution
        --
        Set the genotype of individual INDIV at snp SNP to value VALUE
        Correct Genotype is a value of length 2, with alleles matching the SNP
        '''
        try:
            i=self.indivIdx[indiv]
            j=self.snpIdx[snp]
            self.Data[i,j]=self.snp[snp].addObservation(value)     
        except:
            print( 'Error')
            raise
    def printSNP(self,stream=sys.stdout):
        '''
        Print SNP information
        '''
        print( 'name\tA1\tA2\tcallRate', file = stream)
        for sName in sorted(self.snp,key=lambda x:self.snpIdx[x]):
            geno=self.SnpGenotype(sName)
            s=self.snp[sName]
            print(s.name,s.alleles[0],s.alleles[1],np.mean([g is not missing for g in geno]), file = stream)
    ##### PEDIGREE AND FAMILIES
    def BuildPedigree(self):
        '''
        Construct the pedigree relationships between individuals in the dataset
        Identify founders and parent-offspring links
        '''
        self.pedigree=Pedigree(self.indiv.values())
        self.founders=np.array([0]*self.nindiv,dtype=bool)
        for guy in self.indiv.values():
            founder=True
            if self.pedigree.setFather(guy,guy.fatherID):
                founder=False
            if self.pedigree.setMother(guy,guy.motherID):
                founder=False
            if founder:
                self.pedigree.founders.append(guy)
                self.founders[self.indivIdx[guy.id]]=True
    def BuildFamilies(self):
        '''
        Cluster related individuals in families
        Sets the self.families attribute to a list of the families
        '''
        try:
            ped=self.pedigree
        except AttributeError:
            self.BuildPedigree()
            ped=self.pedigree
        ped.BuildFamilies()
        self.families=ped.families
    def UnrelatedIndividuals(self):
        try:
            ped=self.pedigree
        except AttributeError:
            self.BuildPedigree()
            ped=self.pedigree
        return ped.UnrelatedIndividuals()
    def NuclearFamilies(self):
        '''
        Generator that yields all nuclear families in the dataset (with both parents present)
        '''
        try:
            allFams=self.families
        except AttributeError:
            self.BuildFamilies()
            allFams=self.families
        for fam in allFams:
            for nucFam in fam.NuclearFamilies():
                yield nucFam
    
    def HalfSibFamilies(self):
        '''
        Generator that yields all half-sib families in the dataset.
        '''
        try:
            allFams=self.families
        except AttributeError:
            self.BuildFamilies()
            allFams=self.families
        for fam in allFams:
            for hsFam in fam.HalfSibFamilies():
                yield hsFam
    
    def checkSegregation(self,autosome=True,sex_type='XY'):
        # if not autosome:
        #     if sex_type=='XY':
        #         return self.checkSegregationXY()
        #     elif sex_type=='ZW':
        #         return self.checkSegregationZW()
        try:
            allFams=self.families
        except AttributeError:
            self.BuildFamilies()
            allFams=self.families
        self.MendelErrors={}
        for fam in allFams:
            for node in fam.nonFounders:
                o=node.indiv
                ## father <-> offspring
                if (autosome or sex_type=='ZW') and node.father is not None:
                    father=node.father.indiv
                    for snp in self.AllSnps():
                        go=self.Genotype(o.id,snp)
                        gf=self.Genotype(father.id,snp)
                        if go is missing or gf is missing:
                            continue
                        if (go==0 and gf==2) or (go==2 and gf==0):
                            try:
                                self.SetGenotype(o.id,snp,missing)
                                self.SetGenotype(father.id,snp,missing)
                            except:
                                print( o.id,father.id,go,gf)
                                
                            try :
                                self.MendelErrors[snp]+=1
                            except KeyError:
                                self.MendelErrors[snp]=1
                            try :
                                self.MendelErrors[o.id]+=1
                            except KeyError:
                                self.MendelErrors[o.id]=1
                            try :
                                self.MendelErrors[father.id]+=1
                            except KeyError:
                                self.MendelErrors[father.id]=1
                ## mother <-> offspring
                if (autosome or sex_type=='XY') and node.mother is not None:
                    mother=node.mother.indiv
                    for snp in self.AllSnps():
                        go=self.Genotype(o.id,snp)
                        gm=self.Genotype(mother.id,snp)
                        if go is missing or gm is missing:
                            continue
                        if (go==0 and gm==2) or (go==2 and gm==0):
                            self.SetGenotype(o.id,snp,missing)
                            self.SetGenotype(mother.id,snp,missing)
                            try :
                                self.MendelErrors[snp]+=1
                            except KeyError:
                                self.MendelErrors[snp]=1
                            try :
                                self.MendelErrors[o.id]+=1
                            except KeyError:
                                self.MendelErrors[o.id]=1
                            try :
                                self.MendelErrors[mother.id]+=1
                            except KeyError:
                                self.MendelErrors[mother.id]=1
        return self.MendelErrors
    ### POPULATION GENETICS methods
    def compute_pop_frq(self,verbose=True):
        '''
        Compute allele frequencies for each population in the dataset, considering only
         founders.
        '''
        try:
            fvec=self.founders
        except AttributeError:
            self.BuildPedigree()
            fvec=self.founders
        pop_names=[]
        frqs=[]
        ## identify missing markers
        wtot = np.ones( self.Data.shape[1], dtype=np.bool)
        for pname,pvec in self.populations.items():
            pop_founders = pvec&fvec
            ##w = complete_cases(self.Data[pop_founders,])
            w = self.Data[pop_founders,] != missing
            wbar = np.sum( w, axis = 0)
            nmiss = np.sum(wbar==0)
            ##if np.min( wbar) == 0:
            if nmiss >0 :
                print( '\t%16s: %d individuals, %d SNPs with no info'%(pname, sum(pvec),nmiss))
            wtot &= wbar > 0
        print( '\tFound %d SNPs with info on all pops'%(np.sum(wtot!=0)))
        ## Compute allele frequencies
        for pname,pvec in self.populations.items():
            print('\t %16s'%pname)
            pop_founders = pvec&fvec
            w = complete_cases(self.Data[pop_founders,])
            ff = np.zeros( self.Data.shape[1], dtype=np.float)
            subw = w[:,wtot]
            subdata = self.Data[pop_founders,][:,wtot]
            ##
            ff[ wtot ] = 0.5*np.average( subdata, axis=0, weights=subw)
            ## using mean
            ff[ wtot ] = 0.5*np.mean(subdata*subw,axis=0)/np.mean(subw,axis=0)
            subset = wtot == 0
            ff[ subset ] = np.ma.masked
            frqs.append(ff)
            pop_names.append(pname)
        print()
        return {"pops":tuple(pop_names),'freqs':np.vstack(frqs)}
   
class Family():
    def __init__(self,founders=[],nonFounders=[]):
        self.founders=founders
        self.nonFounders=nonFounders
    def __str__(self):
        tw='Founders: '+'x'.join([x.indiv.id for x in self.founders])+' with '+str(len(self.nonFounders))+' offspring'
        return tw
    def addFounder(self,founder):
        ''' Add a node to the founders '''
        if founder not in self.founders:
            self.founders.append(founder)
    def addNonFounders(self,offspring):
        ''' Add a list of nodes to the offspring '''
        if offspring not in self.nonFounders:
            self.nonFounders.append(offspring)
    def NuclearFamilies(self):
        '''
            Build Nuclear Families within an extended pedigree (family)
            Family founders are ordered so that fam.founders[0] is dad
        '''
        FSfam=[]
        children=list(self.nonFounders)
        while len(children)>0:
            firstFS=children[0]
            papa=firstFS.father
            maman=firstFS.mother
            if papa is not None and maman is not None:
                enfants=[off for off in children if off.father==papa and off.mother==maman]
                FSfam.append(Family(founders=[papa,maman],nonFounders=enfants))
                for off in enfants:
                    children.remove(off)
            else:
                children.remove(firstFS)
        return FSfam
    def _rmHiddenFullSibs(self,enfants,fatherConnected=True):
        '''
        remove "hidden" full sibs from a set of individuals (i.e.
        have same motherID (fatherConnected=True) 
        or fatherID (fatherConnected=False))
        '''
        keepThem=[]
        trashThem=[]
        otherParents={}
        ## make sure we consider enfants always in the same order
        enfants=sorted(enfants,key=lambda x:x.indiv.id)
        for node in sorted(enfants,key=lambda x:x.indiv.callrate,reverse=True):
            guy=node.indiv
            if fatherConnected:
                otherID=guy.motherID
            else:
                otherID=guy.fatherID
            try:
                # we have encountered a full sib before
                # we don't keep that guy
                otherParents[otherID]+=1
                trashThem.append(node)
            except KeyError:
                otherParents[otherID]=1
                keepThem.append(node)
        return keepThem,trashThem
                    
    def HalfSibFamilies(self):
        '''
        Return HalfSib Families contained in the family
        '''
        HSfam=[]
        children=list(self.nonFounders)
        while len(children)>0:
            firstHS=children[0]
            papa=firstHS.father
            maman=firstHS.mother
            if papa is not None:
                if maman is None:
                    ## be sure all of them have none mothers
                    enfants=[off for off in children if off.father==papa and off.mother is None]
                    for off in enfants:
                        children.remove(off)
                    ## remove hidden FS
                    ##enfants,rmvd=self._rmHiddenFullSibs(enfants)
                    HSfam.append(Family(founders=[papa],nonFounders=enfants))
                else:
                    ## nuclear family: remove everyone
                    enfants=[off for off in children if off.father==papa and off.mother==maman]
                    for off in enfants:
                        children.remove(off)
            elif maman is not None:
                ## father is None
                enfants=[off for off in children if off.mother==maman and off.father is None]
                for off in enfants:
                    children.remove(off)
                ## remove hidden FS
                ##enfants,rmvd=self._rmHiddenFullSibs(enfants,fatherConnected=False)
                HSfam.append(Family(founders=[maman],nonFounders=enfants))
            else:
                children.remove(firstHS)
        return HSfam
    
    
class Pedigree():
    ''' Pedigree relationships between individuals'''
    def __init__(self,guys=[]):
        self.nodes={}
        self.founders=[]
        self.families=[]
        for guy in guys:
            self.addNode(guy)
    def addNode(self,guy):
        ''' Add a new guy to the pedigree'''
        self.nodes[guy.id]=PedNode(guy)
    def setFather(self,son,dadID):
        ''' 
        Set that dadId is dad of sonId 
        return True if dad found
        return False if not found
        raise if sonID not found
        '''
        if not son.id in self.nodes:
            print( "Don't know that guy",son.id)
            raise KeyError 
        try:
            self.nodes[son.id].father=self.nodes[dadID]
            self.nodes[dadID].children.append(self.nodes[son.id])
            return True
        except KeyError:
            return False
    def setMother(self,son,momID):
        ''' 
        Set that momId is mom of sonId 
        return True if mom found
        return False if not found
        raise if sonID not found
        '''
        if not son.id in self.nodes:
            print( "Don't know that guy",sonId)
            raise KeyError 
        try:
            self.nodes[son.id].mother=self.nodes[momID]
            self.nodes[momID].children.append(self.nodes[son.id])
            return True
        except KeyError:
            return False    
    def _getRelatives(self,node,connected):
        if node in connected:
            return
        connected.append(node)
        for off in [x for x in node.children]:
            self._getRelatives(off,connected)
        if node.father!=None:
            self._getRelatives(node.father,connected)
        if node.mother!=None and node.mother not in connected:
            self._getRelatives(node.mother,connected)
    def UnrelatedIndividuals(self):
        return [x.id for x in self.founders if len(self.nodes[x.id].children)==0]
    def BuildFamilies(self):
        '''
        Cluster individuals by families
        '''
        founders2go=[self.nodes[x.id] for x in self.founders[:]]
        offspring2go=list(set(self.nodes.values())-(set(founders2go)&set(self.nodes.values())))
        while len(founders2go)>0:
            relatives=[]
            self._getRelatives(founders2go[0],relatives)
            famFounders=[]
            famNonFounders=[]
            for x in relatives:
                if x in founders2go:
                    famFounders.append(x)
                else:
                    famNonFounders.append(x)
            if len(famNonFounders)>0:
                self.families.append(Family(founders=famFounders,nonFounders=famNonFounders))
            for x in famFounders:
                founders2go.remove(x)
            for x in famNonFounders:
                try:
                    offspring2go.remove(x)
                except ValueError:
                    print( 'pula',x.indiv.id)

class PedNode():
    def __init__(self,indiv,father=None,mother=None):
        self.indiv=indiv
        self.father=father
        self.mother=mother
        self.children=[]
    def addChild(self,child):
        if child not in self.children:
            self.children.append(child)
        
class Individual():
    def __init__(self,id,pop=None,sex=None,fatherID=None,motherID=None,phenotype=None):
        self.id=id
        self.pop=pop
        self.sex=sex
        self.fatherID=fatherID
        self.motherID=motherID
        self.phenotype=[phenotype]
        self.callrate=0


class SNP():
    def __init__(self,Name):
        self.name=Name
        self.alleles=[None,None] ## maps 0,1 to alleles
        self.mapping={} ## maps allele to 0,1
        self.counts=np.zeros(3,dtype='int16') ## number(a1),number(a2),number(missing)
        self.Gcounts=[0,0,0]
        self.addObservation=self.addObservationNotFull
    def callrate(self):
        return 0.5*(self.counts[0]+self.counts[1])/(0.5*(self.counts[0]+self.counts[1])+self.counts[2])
    def initAlleles(self,a1,a2):
        self.alleles[0]=a1
        self.alleles[1]=a2
        self.mapping[a1]=0
        self.mapping[a2]=1
        self._create_mapping()
        self.addObservation=self.addObservationFull
    def delGenotype(self,geno):
        '''
        Remove a genotype from observed alleles
        '''
        if geno == missing:
            self.counts[2] -= 1
        else:
            self.counts[0] -= 2-geno
            self.counts[1] -= geno
    def addGenotype(self,geno):
        '''
        Add a genotype to observed alleles
        '''
        if geno==missing:
            self.counts[2]-=1
        else:
            self.counts[0] += 2-geno
            self.counts[1] += geno
    def _create_mapping(self):
        self.g2int={}
        self.g2int[self.alleles[0]+self.alleles[0]]=zero
        self.g2int[self.alleles[0]+self.alleles[1]]=un
        self.g2int[self.alleles[1]+self.alleles[0]]=un
        self.g2int[self.alleles[1]+self.alleles[1]]=deux
    def addObservationFull(self,value):
        try:
            return self.g2int[value]
        except KeyError:
            return missing
    def addObservationNotFull(self,value):
        ''' 
        Add an observed genotype ('AA','01'...) to the SNP 
        Return the 'integer genotype',0,1,2
        if two alleles have already been observed, switch to fast
        version
        '''
        geno=missing
        try:
            i1,i2=value[0],value[1]
        except :
            return geno
        try:
            g1=self.mapping[i1]
        except KeyError:
            if self.alleles[0] is None:
                self.alleles[0]=i1
                self.mapping[i1]=zero
                g1=zero
            elif self.alleles[1] is None:
                self.alleles[1]=i1
                self.mapping[i1]=un
                g1=un
                self._create_mapping()
                self.addObservation=self.addObservationFull
            else:
                return geno
        try:
            g2=self.mapping[i2]
        except KeyError:
            assert self.alleles[0]!=None
            if self.alleles[1] is None:
                self.alleles[1]=i2
                self.mapping[i2]=un
                g2=un
                self._create_mapping()
                self.addObservation=self.addObservationFull
            else:
                return geno
        geno=g1+g2
        return geno

class Locus():
    def __init__(self,nom,position):
        self.name=nom
        self.pos=position
    def __eq__(self,other):
        return self.pos == other.pos
    def __ne__(self,other):
        return self.pos != other.pos
    def __lt__(self,other):
        return self.pos < other.pos
    def __le__(self,other):
        return self.pos <= other.pos
    def __gt__(self,other):
        return self.pos > other.pos
    def __ge__(self, other):
        return self.pos >= other.pos

class Map():
    def __init__(self):
        '''
        Map class contains maps (duh)
        The chromosome names are stored in the chromosomes attribute (in sorted order)
        Each chromosome has two maps (genetic/RH and physical)
        '''
        self.chromosomes=[] 
        self.cartes={}
        self.markers={}
        self.input_order=[]
    def write(self,stream=sys.stdout):
        for chrom in self.chromosomes:
            for mk in self.physMap(chrom):
                pos=self.position(mk)
                print( '\t'.join([str(x) for x in [pos[0],mk,int(pos[1]),int(pos[2])]]), file=stream)
        return
    def addMarker(self,M,C,posG=0,posP=0):
        '''
        Add a marker on chromosome C, with genetic position posG and
        physical position posP
        '''
        self.input_order.append(M)
        if not C in self.cartes:
            insort(self.chromosomes,C)
            self.cartes[C]={'Genet':[],'Phys':[]}
        locG=Locus(M,posG)
        locM=Locus(M,posP)
        self.markers[M]=(C,locG,locM)
        insort(self.cartes[C]['Genet'],locG)
        insort(self.cartes[C]['Phys'],locM)
    def delMarker(self,M):
        try:
            locus = self.markers[M]
        except KeyError:
            print( "No such marker")
        del self.markers[M]
        self.cartes[locus[0]]['Genet'].remove(locus[1])
        self.cartes[locus[0]]['Phys'].remove(locus[2])
    def genetMap(self,C,xleft=-1,xright=np.inf):
        '''
        Returns and iterator on the genetic map between position xleft and xright 
        on chromosome C
        '''
        try:
            carte=self.cartes[C]['Genet']
        except KeyError:
            return []
        mk=carte[bisect_left(carte,Locus(0,xleft)):bisect_right(carte,Locus(0,xright))]
        return (m.name for m in mk)
    def physMap(self,C,xleft=0,xright=np.inf):
        '''
        Returns and iterator on the genetic map between position xleft and xright 
        on chromosome C
        '''
        try:
            carte=self.cartes[C]['Phys']
        except KeyError:
            return []
        mk=carte[bisect_left(carte,Locus(0,xleft)):bisect_right(carte,Locus(0,xright))]
        return (m.name for m in mk)
    def length(self,C):
        '''
        Return the length of maps of chromosome C [physical,genetic]
        '''
        carte=self.cartes[C]['Phys']
        lenP=carte[-1].pos-carte[0].pos
        carte=self.cartes[C]['Genet']
        lenG=carte[-1].pos-carte[0].pos
        return [lenP,lenG]
    def position(self,M):
        '''
        Returns chromosome,genetic,physical position of marker M
        '''
        try:
            locus=self.markers[M]
            return locus[0],locus[1].pos,locus[2].pos
        except KeyError:
            print( M,'not found')
            return None
    def sort_loci(self,loci,physical=True):
        '''
        Sort loci according to their position on the map

        Parameters:
        ---------------

        loci : a list of loci names
        physical : if True sort according to physical position, o.w. use genetic  / RH position

        Return Value:
        -----------------

        list of the same loci, sorted
        '''
        new_list=[]
        for locus in loci:
            try:
                tmp=self.markers[locus]
                if physical:
                    myloc=tmp[2]
                else:
                    myloc=tmp[1]
                try:
                    mychr=int(tmp[0])
                except:
                    mychr=tmp[0]
            except KeyError:
                continue
            new_list.append((locus,mychr,myloc.pos))
        new_list=sorted(new_list,key=itemgetter(1,2))
        return [x[0] for x in new_list]
    
    def iterate_pos_windows(self,wlen,padlen=0):
        for c in self.chromosomes:
            pos_left=0
            pos_right=-1
            clen=self.length(c)[0]
            print( c,clen)
            while pos_right < clen:
                pos_right=pos_left+wlen
                yield (c,(pos_left-padlen)>0 and pos_left-padlen or 0,(pos_right+padlen)>clen and clen+1 or pos_right+padlen )
                pos_left=pos_right
                



import sys
import re
import argparse
import numpy as np
from bisect import insort
from hapflk import missing
from hapflk import data
import hapflk._pgenlib as _pgenlib

def plink_bfile(string):
    try:
        open(string+'.bed')
    except IOError:
        msg="can't open BED file %s"%string+'.bed'
        raise argparse.ArgumentTypeError(msg)
    try:
        open(string+'.fam')
    except IOError:
        msg="can't open FAM file %s"%string+'.fam'
        raise argparse.ArgumentTypeError(msg)
    try:
        open(string+'.bim')
    except IOError:
        msg="can't open BIM file %s"%string+'.bim'
        raise argparse.ArgumentTypeError(msg)
    return string

def shapeit_file(string):
    try:
        open(string+'.haps')
    except IOError:
        msg="can't open HAPS file%s"%string+'.haps'
        raise argparse.ArgumentTypeError(msg)
    try:
        open(string+'.sample')
    except IOError:
        msg="can't open SAMPLE file%s"%string+'.sample'
        raise argparse.ArgumentTypeError(msg)
    return string

io_parser=argparse.ArgumentParser(add_help=False,formatter_class=argparse.ArgumentDefaultsHelpFormatter)
input_group=io_parser.add_argument_group('Input Files')
input_type_group = input_group.add_mutually_exclusive_group(required=True)
input_type_group.add_argument('--bfile',metavar='PREFIX',help='PLINK bfile prefix (bim,fam,bed)',type=plink_bfile)
input_type_group.add_argument('--sfile',metavar='PREFIX',help='ShapeIT file prefix (haps,sample)',type=shapeit_file)


defaultParams= {
    "MissGeno":'0',
    "MissPheno":'-9',
    "MissParent":'0',
    "MissSex":'0'
    }


key="kshuwewekiuvf"

def _get_snpIdx(myMap,chrom,left,right,othermap=False):
    '''
    Return index of SNPs in map

    Parameters
    --------------
    myMap : the map
    chrom : chromosome
    left : leftmost position
    right : rightmost position
    othermap : use other (genetic / RH) map scale
    '''
    if othermap:
        getMap=myMap.genetMap
    else:
        getMap=myMap.physMap
    mySnps=[ x for x in getMap(chrom,xleft=left,xright=right)]
    mySnpIdx=[]
    for i,s in enumerate(myMap.input_order):
        if s in mySnps:
            mySnpIdx.append(i)
    return mySnpIdx

##### generic input function ####
def parseInput(options,params=defaultParams):
    '''
    Read input files 
    
    Parameters
    --------------
    options : options from io_parser
    params : parameters for coding used in input files, overidden using options

    Returns
    ----------
    res: dictionary with keys:
         -- dataset : an instance from the gwas Dataset class
         -- map : an instance from the gwas Map class

    See also
    ----------
    data.Dataset
    data.Map
    '''

    ## read input
    if options.bfile:
        return parsePlinkBfile(options.bfile,params=params,options=options)
    elif options.sfile:
        return parseShapeIT(options.sfile, params=params)
        
               
################ SHAPEIT2 files ####################################
def parseShapeIT(prefix, params=defaultParams):
    ## parse sample file
    idata=[]
    with open(prefix+'.sample') as f:
        for il,ligne in enumerate(f):
            if il<2:
                continue
            buf=ligne.split()
            phe=((buf[6]!=params['MissPheno']) and float(buf[6])) or None
            idata.append([buf[0], \
                          buf[1],  \
                              ((buf[3]!=params['MissParent']) and buf[3]) or None, \
                              ((buf[4]!=params['MissParent']) and buf[4]) or None, \
                              (buf[5]!=params['MissSex'] and buf[5]) or None, \
                              phe])
                              
    ## parse haps file
    SNPs=[]
    myMap=data.Map()
    hdata = []
    with open(prefix+'.haps') as f:
        for ligne in f:
            buf=ligne.split()
            hdata.append( [ int(i) for i in buf[5:] ] )
            ## create SNP
            myS=data.SNP(buf[1])
            myS.alleles[0]=buf[3]
            myS.alleles[1]=buf[4]
            ## add marker to map
            try:
                mychr=int(buf[0])
            except:
                mychr=buf[0]
            mychr=buf[0]
            myMap.addMarker(M=buf[1],C=mychr,posG=float(buf[2])*1e-6,posP=float(buf[2]))
            SNPs.append(myS)
    sdata = {'snps':SNPs, 'map':myMap}
    ni = 2*len(idata)
    ns = len(sdata['snps'])
    mySnpIdx = range(ns)
    dataset = data.Dataset(prefix,nsnp = ns, nindiv = ni)
    ## fill in SNP data
    for s in [sdata['snps'][i] for i in mySnpIdx]:
        dataset.addSnp(s.name)
        dataset.snp[s.name].initAlleles(s.alleles[0],s.alleles[1])
    for iind,ind in enumerate(idata):
        dataset.addIndividual(pop=ind[0],
                              ID=ind[1]+'.1',
                              fatherID=ind[2],
                              motherID=ind[3],
                              sex=ind[4],
                              phenotype=ind[5])
        dataset.addIndividual(pop=ind[0],
                              ID=ind[1]+'.2',
                              fatherID=ind[2],
                              motherID=ind[3],
                              sex=ind[4],
                              phenotype=ind[5])
        dataset.Data[2*iind  ,:]=np.array( [ 2*hdata[s][2*iind] for s in mySnpIdx], dtype=np.int)
        dataset.Data[2*iind+1,:]=np.array( [ 2*hdata[s][2*iind+1] for s in mySnpIdx], dtype=np.int)
    return {'dataset':dataset, 'map': sdata['map']}

################ BED / BIM / FAM files #############################

def parsePlinkBfile(prefix,noPheno=False,params=defaultParams,options=None):
    fam=prefix+'.fam'
    bim=prefix+'.bim'
    bed=prefix+'.bed'
    ## individual data
    idata=parseFamFile(fam) ## options not implemented
    ni=len(idata)
    ### SNP data
    sdata=parseBimFile(bim)
    ns=len(sdata['snps'])
    mySnpIdx=range(ns)
    dataset=data.Dataset(prefix,nsnp=ns,nindiv=ni)
    ## fill in SNP data
    for s in [sdata['snps'][i] for i in mySnpIdx]:
        dataset.addSnp(s.name)
        dataset.snp[s.name].initAlleles(s.alleles[0],s.alleles[1])
    ## fill in indiv data
    for ind in idata:
        dataset.addIndividual(pop=ind[0],
                              ID=ind[1],
                              fatherID=ind[2],
                              motherID=ind[3],
                              sex=ind[4],
                              phenotype=ind[5])
    fillBedData_fast(bed.encode(),dataset.Data)
    return {'dataset':dataset,'map':sdata['map']}

def parseFamFile(fileName,noPheno=False,params=defaultParams):
    indiv_data=[]
    with open(fileName) as f:
        for ligne in f:
            buf=ligne.split()
            if noPheno:
                phe=None
            else:
                phe=((buf[5]!=params['MissPheno']) and float(buf[5])) or None
  
            indiv_data.append([buf[0], \
                               buf[1],  \
                               ((buf[2]!=params['MissParent']) and buf[2]) or None, \
                               ((buf[3]!=params['MissParent']) and buf[3]) or None, \
                               (buf[4]!=params['MissSex'] and buf[4]) or None, \
                                phe])
    return indiv_data

def parseBimFile(fileName):
    SNPs=[]
    myMap=data.Map()
    with open(fileName) as f:
        for ligne in f:
            buf=ligne.split()
            ## create SNP
            myS=data.SNP(buf[1])
            myS.alleles[0]=buf[4]
            myS.alleles[1]=buf[5]
            ## add marker to map
            try:
                mychr=int(buf[0])
            except:
                mychr=buf[0]
            mychr=buf[0]
            myMap.addMarker(M=buf[1],C=mychr,posG=float(buf[2]),posP=float(buf[3]))
            SNPs.append(myS)
    return {'snps':SNPs,'map':myMap}

def fillBedData_fast(fileName,DataMatrix):
    n_indiv,n_snp=DataMatrix.shape
    reader=_pgenlib.PgenReader(fileName,raw_sample_ct=n_indiv)
    buf=np.empty(n_indiv,np.int8)
    for isnp in range(n_snp):
        reader.read(isnp,buf,0)
        buf[buf==-9]=missing
        DataMatrix[:,isnp]=buf

def fillBedData_slow(fileName,DataMatrix,snpidx=None):
    n_indiv,n_snp=DataMatrix.shape
    ## use dict type for faster look up
    ## key index in total, val : index in DataMatrix
    if snpidx is None:
        snpidx=dict([ (i,i) for i in range(n_snp)])
    else:
        snpidx=dict([ (s,i) for i,s in enumerate(snpidx)])
    cur_snp=0
    ## bit enumerator
    def bits(f):
        bytes = (ord(b) for b in f.read())
        for b in bytes:
            for i in range(8):
                yield (b >> i) & 1
    ## bit pair to geno converter
    def bpair_2_geno(pair):
        '''
        from plink doc
        '''
        if pair=='00':
            return 0
        if pair=='11':
            return 2
        if pair=='01':
            return 1
        if pair=='10':
            return missing
    ## constants
    target_magic='0011011011011000'
    snp_major='10000000'    
    indiv_major='00000000'
    ### let's go
    bed_stream=open(fileName, 'r')
    magic=[]
    f_mode=[]
    i_major=0
    i_minor=0
    proceed_2_next_bit=False
    bit_geno=[]
    nbit=-1
    for b in bits(bed_stream):
        nbit+=1
        ## Test for magic number
        if nbit<16:
            magic.append(str(b))
            continue
        if nbit==16:
            magic=''.join(magic)
            if magic!=target_magic:
                print('Not a bed file')
                raise ValueError
        ## Find out packing mode
        if nbit<24:
            f_mode.append(str(b))
            continue
        if nbit==24:
            f_mode=''.join(f_mode)
            if f_mode==snp_major:
                pack_len=n_indiv 
            elif f_mode==indiv_major:
                pack_len=n_snp
            else:
                print('Cannot determine packing mode, abort')
                raise ValueError
        ## Now i>24, we are reading genotypes
        ## case where we just wait for next bit to start reading again
        if proceed_2_next_bit:
            if nbit%8!=0:
                continue
            else:
                proceed_2_next_bit=False
        ## reading genotypes        
        bit_geno.append(str(b))
        if nbit%2!=0:
            ##assert len(bit_geno)==2
            ## print nbit,i_major,i_minor,bit_geno
            geno=bpair_2_geno(''.join(bit_geno))
            if f_mode==snp_major:
                try:
                    data_idx=snpidx[i_major]
                    DataMatrix[i_minor,data_idx]=geno
                except KeyError:
                    pass
            else:
                try:
                    data_idx=snpidx[i_minor]
                    DataMatrix[i_major,data_idx]=geno
                except KeyError:
                    pass
            i_minor += 1 
            bit_geno=[]
        if i_minor==pack_len:
            proceed_2_next_bit=True
            i_minor=0
            i_major+=1
        

############################ Covariate Information Files ##################

## File Format

## Header :ID cov1_name cov2_name ...
## Indiv1/pop1 val1 val2 ...

## add parser options to read in covariate files
def populate_covariate_args(parser):
    ''' Add options to read covariates in option parser'''
    covariate_args=parser.add_argument_group('Covariates','')
    covariate_args.add_argument('--covar',dest='covar_file',metavar='FILE',default=None,
                                help='File with available covariates. Use --cov or --qcov to include them in the analysis')
    covariate_args.add_argument('--cov',metavar='LIST',dest='cov',default='',
                                help='List of *qualitative* covariates to include, separated by commas')
    covariate_args.add_argument('--qcov',metavar='LIST',dest='qcov',default='',
                                help='List of *quantitative*  covariates to include, separated by commas')

def get_covariates_raw(fichier,missing_value='NA'):
    ''' Get raw data from covariate file /fichier/

    File Format:
    Header     : ID cov1_name cov2_name ...
    Other lines: ID1 val1_1 val1_2 ...

    Returns a dictionary with covariate names as keys (read from header)
    For each covariate, the value is a dictionary with IDs as keys and string as value
    No conversion is done. Missing value key can be specified (default = NA)
    '''
    f=open(fichier)
    covar_dat=f.readlines()
    covar_name=covar_dat[0].split()[1:]
    covariates={}
    for nom in covar_name:
        covariates[nom]={}
    for ligne in covar_dat[1:]:
        buf=ligne.split()
        indiv=buf[0]
        for i,covar in enumerate(covar_name):
            if buf[i+1]==missing_value:
                continue
            covariates[covar][indiv]=buf[i+1]## no conversion done here
    return covariates

def get_covariates_matrix(filename,fact_names,qcov_names,names,stdize=True):
    ''' 
    Returns a design matrix of all covariates listed in fact_names and qcov_names. 
    fact_names, qcov_names: string of qualitative (resp. quantitative) covariates, separated by commas.
    names : list of ID names providing the required ordering of rows.
    stdize : if True, standardize quantitative covariates (X-m)/s.

    Returns a dictionary with keys:
    -- DesignMatrices: dict with covar names as keys and DM as value
    -- levels: dict with cofact names as keys and ordered levels as value
    '''
    covariates=get_covariates_raw(filename)
    cofact=[] ## qualitative cofactors
    covar=[] ## quantitative covariables
    nids=len(names)
    try:
        for c in fact_names.split(','):
            if c!='':
                if c in covariates:
                    cofact.append(c)
                else:
                    print( 'Covariate %s not found in covariate file, ignoring.'%c)
    except:
        pass
    try:
        for c in qcov_names.split(','):
            if c!='':
                if c in covariates:
                    covar.append(c)
                else:
                    print( 'Covariate %s not found in covariate file, ignoring.'%c)
    except:
        pass
    Matrices={}
    levels={}
    for q in cofact:
        q_vec=[]
        for nom in names:
            q_vec.append(covariates[q][nom])
        u,aa=np.unique(q_vec,return_inverse=True)
        print( 'Cofactor "%s" with %d levels'%(q,len(u)))
        print( 'Levels:' *u)
        levels[q]=u
        M=np.zeros((nids,len(u)),dtype=np.int)
        M[range(nids),aa]=1
        Matrices[q]=M[:,1:]
    for q in covar:
        M=np.zeros((nids,1),dtype=np.float)
        for i,nom in enumerate(names):
            try:
                val=float(covariates[q][nom])
                M[i,0]=val
            except ValueError:
                raise ValueError('Covariate %s is not quantitative !'%q)
        mu=np.mean(M)
        ss=np.std(M)
        if stdize:
            M=(M-mu)/ss
        print( 'Covariate "%s" with mean %f and sd %f'%(q,mu,ss))
        Matrices[q]=M
    return { 'cofactors' : cofact,
                 'covariables' : covar,
                 "DesignMatrices":Matrices,
                 "factor_levels":levels}
